#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a taurus text editor widget."""

__all__ = ["TaurusBaseEditor"]

__docformat__ = 'restructuredtext'

import sys
import os

from taurus.external.qt import Qt

try:
    import spyderlib
    v = spyderlib.__version__.split('.', 2)[:2]
    v = map(int, v)
    if v < [2, 1]:
        raise Exception("TaurusEditor needs spyderlib >= 2.1")
except ImportError:
    raise Exception("TaurusEditor needs spyderlib >= 2.1")

from spyderlib.utils.qthelpers import create_toolbutton
from spyderlib.widgets.findreplace import FindReplace
from spyderlib.widgets.editortools import OutlineExplorerWidget
from spyderlib.widgets.editor import EditorMainWindow, EditorSplitter
from taurus.qt.qtgui.util import ActionFactory


class TaurusBaseEditor(Qt.QSplitter):

    def __init__(self, parent=None):
        Qt.QSplitter.__init__(self, parent)

        self.editorstacks = []
        self.editorwindows = []

        self.menu_actions, self.io_actions = self.createMenuActions()

        self.toolbar_list = None
        self.menu_list = None

        self.find_widget = FindReplace(self, enable_replace=True)
        self.outlineexplorer = OutlineExplorerWidget(self, show_fullpath=False,
                                                     show_all_files=False)
        editor_widgets = Qt.QWidget(self)
        editor_layout = Qt.QVBoxLayout()
        editor_layout.setContentsMargins(0, 0, 0, 0)
        editor_widgets.setLayout(editor_layout)

        editor_layout.addWidget(EditorSplitter(self, self, self.menu_actions,
                                               first=True))
        editor_layout.addWidget(self.find_widget)

        self.setContentsMargins(0, 0, 0, 0)
        self.addWidget(editor_widgets)
        self.addWidget(self.outlineexplorer)

        self.setStretchFactor(0, 1)
        self.setStretchFactor(1, 0)

        self.setup_window([], [])

    def editorStack(self):
        return self.editorstacks[0]

    def createMenuActions(self):
        return [], []

    def closeEvent(self, event):
        for win in self.editorwindows[:]:
            win.close()
        event.accept()

    def load(self, filename, goto=None):
        editorstack = self.editorStack()
        fileinfo = editorstack.load(filename)
        editorstack.analyze_script()
        if goto is not None:
            fileinfo.editor.go_to_line(goto)

    def reload(self, idx=None, filename=None, goto=None):
        if idx is None:
            idx = is_file_opened(filename)
        if idx is not None:
            editorstack = self.editorStack()
            editorstack.reload(idx)
            editorstack.analyze_script()
            if goto is not None:
                editorstack.set_current_filename(filename)
                editorstack.data[idx].editor.go_to_line(goto)

    def set_current_filename(self, filename):
        self.editorStack().set_current_filename(filename)

    def register_editorstack(self, editorstack):
        self.editorstacks.append(editorstack)
        if self.isAncestorOf(editorstack):
            # editorstack is a child of the Editor plugin
            editorstack.set_fullpath_sorting_enabled(True)
            editorstack.set_closable(len(self.editorstacks) > 1)
            editorstack.set_outlineexplorer(self.outlineexplorer)
            editorstack.set_find_widget(self.find_widget)
            oe_btn = create_toolbutton(self)
            oe_btn.setDefaultAction(self.outlineexplorer.visibility_action)
            editorstack.add_corner_widgets_to_tabbar([5, oe_btn])

        editorstack.set_io_actions(*self.io_actions)
        font = Qt.QFont("Monospace")
        font.setPointSize(10)
        editorstack.set_default_font(font, color_scheme='Spyder')
        editorstack.close_file.connect(self.close_file_in_all_editorstacks)
        editorstack.create_new_window.connect(self.create_new_window)
        editorstack.plugin_load.connect(self.load)

    def unregister_editorstack(self, editorstack):
        self.editorstacks.pop(self.editorstacks.index(editorstack))

    def clone_editorstack(self, editorstack):
        editorstack.clone_from(self.editorStack())

    def setup_window(self, toolbar_list, menu_list):
        self.toolbar_list = toolbar_list
        self.menu_list = menu_list

    def create_new_window(self):
        window = EditorMainWindow(self, self.menu_actions,
                                  self.toolbar_list, self.menu_list,
                                  show_fullpath=False, fullpath_sorting=True,
                                  show_all_files=False, show_comments=True)
        window.resize(self.size())
        window.show()
        self.register_editorwindow(window)
        window.destroyed.connect(lambda win=window: self.unregister_editorwindow(win))

    def register_editorwindow(self, window):
        self.editorwindows.append(window)

    def unregister_editorwindow(self, window):
        self.editorwindows.pop(self.editorwindows.index(window))

    def get_focus_widget(self):
        pass

    def close_file_in_all_editorstacks(self, index):
        sender = self.sender()
        for editorstack in self.editorstacks:
            if editorstack is not sender:
                editorstack.blockSignals(True)
                editorstack.close_file(index)
                editorstack.blockSignals(False)

    def register_widget_shortcuts(self, context, widget):
        """Fake!"""
        pass

    def refresh_save_all_action(self):
        pass

from spyderlib.plugins.editor import Editor


class TaurusBaseEditor2(Qt.QMainWindow):

    def __init__(self, parent=None):
        Qt.QMainWindow.__init__(self, parent)

        self._editor = Editor(self)

        self.setCentralWidget(self._editor)

    def register_shortcut(self, qaction_or_qshortcut, context, name,
                          default=None):
        pass

    def show_hide_project_explorer(self):
        pass


def demo():
    test = TaurusBaseEditor()
    test.resize(1000, 800)
    test.load(__file__)
    test.load("__init__.py")
    test.show()
    return test


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <file names>"
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Taurus text editor demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()

    if len(args) == 0:
        w = demo()
    else:
        w = TaurusEditor()
        w.resize(900, 800)
        for name in args:
            w.load(name)
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains the base taurus operation classes"""

__all__ = ["TaurusOperation", "WriteAttrOperation"]

__docformat__ = "restructuredtext"

from .util.log import Logger


class TaurusOperation(Logger):

    def __init__(self, name='TaurusOperation', parent=None, callbacks=None):
        self.call__init__(Logger, name, parent)
        if callbacks is None:
            callbacks = []
        self._callbacks = callbacks
        self._dangerMessage = None
        self._isDangerous = False

    def getDevice(self):
        pass

    def setCallbacks(self, callbacks):
        self._callbacks = callbacks

    def getCallbacks(self):
        return self._callbacks

    def execute(self):
        for f in self._callbacks:
            f(operation=self)

    def isDangerous(self):
        return self._isDangerous

    def setDangerMessage(self, dangerMessage=None):
        '''if dangerMessage is None, the operation is considered safe'''
        self._dangerMessage = dangerMessage
        self._isDangerous = dangerMessage is not None

    def getDangerMessage(self):
        return self._dangerMessage

    def resetDangerMessage(self):
        self.setDangerMessage(None)


class WriteAttrOperation(TaurusOperation):

    def __init__(self, attr, value, callbacks=None):
        self.call__init__(TaurusOperation, 'WriteAttrOperation',
                          attr, callbacks=callbacks)
        self.attr = attr
        self.value = value

    def execute(self):
        self.attr.write(self.value)
        TaurusOperation.execute(self)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""
configuration file for an example of how to construct a GUI based on TaurusGUI

This configuration file determines the default, permanent, pre-defined
contents of the GUI. While the user may add/remove more elements at run
time and those customizations will also be stored, this file defines what a
user will find when launching the GUI for the first time.
"""

#==============================================================================
# Import section. You probably want to keep this line. Don't edit this block
# unless you know what you are doing
from taurus.qt.qtgui.taurusgui.utils import PanelDescription, ExternalApp, ToolBarDescription, AppletDescription
# (end of import section)
#==============================================================================


#=========================================================================
# General info.
#=========================================================================
GUI_NAME = 'EXAMPLE 01'
ORGANIZATION = 'Taurus'

#=========================================================================
# Specific logo. It can be an absolute path,or relative to the app dir or a
# resource path. If commented out, ":/taurus.png" will be used
#=========================================================================
# CUSTOM_LOGO = <path GUI-specific logo

#=========================================================================
# You can provide an URI for a manual in html format
# (comment out or make MANUAL_URI=None to skip creating a Manual panel)
#=========================================================================
MANUAL_URI = 'http://www.taurus-scada.org'

#=========================================================================
# If you want to have a main synoptic panel, set the SYNOPTIC variable
# to the file name of a jdraw file. If a relative path is given, the directory
# containing this configuration file will be used as root
# (comment out or make SYNOPTIC=None to skip creating a synoptic panel)
#=========================================================================
SYNOPTIC = ['images/example01.jdw', 'images/syn2.jdw']

#=========================================================================
# Set INSTRUMENTS_FROM_POOL to True for enabling auto-creation of
# instrument panels based on the Pool Instrument info
#=========================================================================
INSTRUMENTS_FROM_POOL = False

#=========================================================================
# Define panels to be shown.
# To define a panel, instantiate a PanelDescription object (see documentation
# for the gblgui_utils module)
#=========================================================================

nxbrowser = PanelDescription('NeXus Browser',
                             classname='TaurusNeXusBrowser',
                             area=None)

i0 = PanelDescription('BigInstrument',
                      classname='TaurusAttrForm',
                      model='sys/tg_test/1')

i1 = PanelDescription('instrument1',
                      classname='TaurusForm',
                      model=['sys/tg_test/1/double_scalar',
                             'sys/tg_test/1/short_scalar_ro',
                             'sys/tg_test/1/float_spectrum_ro',
                                'sys/tg_test/1/double_spectrum'])

i2 = PanelDescription('instrument2',
                      classname='TaurusForm',
                      model=['sys/tg_test/1/wave',
                             'sys/tg_test/1/boolean_scalar'])

trend = PanelDescription('Trend',
                         classname='TaurusTrend',
                         model=['sys/tg_test/1/double_scalar'])

connectionDemo = PanelDescription('Selected Instrument',
                                  classname='PyQt4.Qt.QLineEdit',
                                  sharedDataRead={
                                      'SelectedInstrument': 'setText'},
                                  sharedDataWrite={'SelectedInstrument': 'textEdited'})

#=========================================================================
# Define custom toolbars to be shown. To define a toolbar, instantiate a
# ToolbarDescription object (see documentation for the gblgui_utils module)
#=========================================================================

dummytoolbar = ToolBarDescription('Empty Toolbar',
                                  classname='QToolBar',
                                  modulename='PyQt4.Qt')

# panictoolbar = ToolBarDescription('Panic Toolbar',
#                        classname = 'PanicToolbar',
#                        modulename = 'tangopanic')

#=========================================================================
# Define custom applets to be shown in the applets bar (the wide bar that
# contains the logos). To define an applet, instantiate an AppletDescription
# object (see documentation for the gblgui_utils module)
#=========================================================================

mon2 = AppletDescription('Dummy Monitor',
                         classname='TaurusMonitorTiny',
                         model='eval:1000*rand(2)')


#=========================================================================
# Define which External Applications are to be inserted.
# To define an external application, instantiate an ExternalApp object
# See TaurusMainWindow.addExternalAppLauncher for valid values of ExternalApp
#=========================================================================
xterm = ExternalApp(cmdargs=['xterm', 'spock'],
                    text="Spock", icon='utilities-terminal')
hdfview = ExternalApp(["hdfview"])
pymca = ExternalApp(['pymca'])

#=========================================================================
# Macro execution configuration
# Comment out or make MACRO_SERVER=None or set MACRO_PANELS=False to skip
# creating a macro execution infrastructure.
# Give empty strings if you want to select the values manually in the GUI
#=========================================================================
# MACROSERVER_NAME =
# DOOR_NAME =
# MACROEDITORS_PATH =

#=========================================================================
# Monitor widget (This is *obsolete* now, you can get the same result defining a
# custom applet with classname='TaurusMonitorTiny')
#=========================================================================
# MONITOR = ['sys/tg_test/1/double_scalar_rww']

#=========================================================================
# Adding other widgets to the catalog of the "new panel" dialog.
# pass a tuple of (classname,screenshot)
# -classname may contain the module name.
# -screenshot can either be a resource URL, a file name (either relative to
# the application dir or with an absolute path) or None
#=========================================================================
EXTRA_CATALOG_WIDGETS = [('PyQt4.Qt.QLineEdit', 'logos:taurus.png'),  # a resource
                         ('PyQt4.Qt.QSpinBox', 'images/syn2.jpg'),  # relative
                         #('PyQt4.Qt.QTextEdit','/tmp/foo.png'),  # absolute
                         ('PyQt4.Qt.QLabel', None)]                # none

#=========================================================================
# Define one or more embedded consoles in the GUI.
# Possible items for console are 'ipython', 'tango', 'spock'
# Note: This is still experimental
#=========================================================================
#CONSOLE = ['tango']



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides an arrow based widget."""

__all__ = ["QWheelEdit"]

__docformat__ = 'restructuredtext'

import os
import math
import numpy

from taurus.external.qt import Qt
from taurus.external.pint import Q_

class _ArrowButton(Qt.QPushButton):
    """Private class to be used by QWheelEdit for an arrow button"""

    ArrowPixmapName = "extra_icons:arrow01.svg"

    ButtonSize = 14
    IconSize = ButtonSize - 2

    def __init__(self, id, parent=None):
        Qt.QPushButton.__init__(self, parent)
        self.setAutoDefault(False)
        pixmap = self.getPixmap()
        self.setIcon(Qt.QIcon(pixmap))
        self.setIconSize(
            Qt.QSize(_ArrowButton.IconSize, _ArrowButton.IconSize))
        self.setFocusPolicy(Qt.Qt.ClickFocus)
        self.setFlat(True)
        self.setStyleSheet('_ArrowButton { border: 0px; }')
        self.setMaximumSize(_ArrowButton.ButtonSize, _ArrowButton.ButtonSize)
        #self.setMinimumSize(_ArrowButton.ButtonSize, _ArrowButton.ButtonSize)
        self._id = id
        self._inc = math.pow(10, id)


class _UpArrowButton(_ArrowButton):
    """Private class to be used by QWheelEdit for an up arrow button"""

    ArrowPixmapKey = "upArrow01"

    def __init__(self, id, parent=None):
        _ArrowButton.__init__(self, id, parent)

    def getPixmap(self):
        pm = Qt.QPixmapCache.find(_UpArrowButton.ArrowPixmapKey)
        if pm is None:
            pm = Qt.QPixmap(self.ArrowPixmapName)
            Qt.QPixmapCache.insert(_UpArrowButton.ArrowPixmapKey, pm)
        return pm


class _DownArrowButton(_ArrowButton):
    """Private class to be used by QWheelEdit for a down arrow button"""

    ArrowPixmapKey = "downArrow01"

    def __init__(self, id, parent=None):
        _ArrowButton.__init__(self, id, parent)
        self._inc = -self._inc

    def getPixmap(self):
        pm = Qt.QPixmapCache.find(_DownArrowButton.ArrowPixmapKey)
        if pm is None:
            pm = Qt.QPixmap(self.ArrowPixmapName)
            pm = pm.transformed(Qt.QMatrix().rotate(180))
            Qt.QPixmapCache.insert(_DownArrowButton.ArrowPixmapKey, pm)
        return pm

class _DigitLabel(Qt.QLabel):
    """A private single digit label to be used by QWheelEdit widget"""

    PixmapKeys = map(str, xrange(10)) + ['blank', 'minus', 'point']

    def __init__(self, lbl, parent=None):
        Qt.QLabel.__init__(self, parent)
        self.setAlignment(Qt.Qt.AlignCenter)
        self.setFocusPolicy(Qt.Qt.StrongFocus)
        self.setStyleSheet(
            """QLabel:focus {background-color: rgb(180,180,255); border-width: 1px; border-style: solid; border-color: rgb(200,200,255);}""")
        self.resetSkin()
        self.setText(lbl)
        self._upButton = None
        self._downButton = None

    def setButtons(self, up, down):
        self._upButton = up
        self._downButton = down

    def keyPressEvent(self, key_event):
        if key_event.key() == Qt.Qt.Key_Up:
            self._upButton.click()
        elif key_event.key() == Qt.Qt.Key_Down:
            self._downButton.click()
        elif key_event.key() == Qt.Qt.Key_Right:
            self.focusNextChild()
        elif key_event.key() == Qt.Qt.Key_Left:
            self.focusPreviousChild()
        Qt.QLabel.keyPressEvent(self, key_event)

    def _update_skin_cache(self):
        skin = os.path.join('digits', self.getSkin())

        for i, d in enumerate(self.PixmapKeys):
            full_name = os.path.join(skin, d)
            pm = Qt.QPixmapCache.find(full_name)
            if pm is None:
                file_name = '%s.png' % full_name
                pm = Qt.QPixmap(file_name)
                if pm is None:
                    continue
                if not pm.isNull():
                    Qt.QPixmapCache.insert(full_name, pm)

    def setSkin(self, skin_name):
        self._skin = skin_name
        self._update_skin_cache()
        self.update()

    def getSkin(self):
        return self._skin

    def resetSkin(self):
        self.setSkin('Nixie')

    def setText(self, t):
        k = t
        if k == '.':
            k = 'point'
        elif k == '-':
            k = 'minus'
        elif k is None or k == '+':
            k = 'blank'

        k = os.path.join('digits', self.getSkin(), k)
        pm = Qt.QPixmapCache.find(k)
        if pm is None:
            Qt.QLabel.setText(self, t)
            return
        Qt.QLabel.setText(self, '')
        self.setPixmap(pm)

    skin = Qt.pyqtProperty('QString', getSkin, setSkin, resetSkin)


class _NumericEditor(Qt.QLineEdit):
    """A private editor to be used by QWheelEdit widget"""

    def __init__(self, parent=None):
        Qt.QLineEdit.__init__(self, parent)
        self.setValidator(Qt.QDoubleValidator(self))
        self.setFrame(False)


class QWheelEdit(Qt.QFrame):
    """A widget designed to handle numeric scalar values. It allows interaction
    based on single digit as well as normal value edition."""

    numberChanged = Qt.pyqtSignal(float)
    numberEdited = Qt.pyqtSignal(float)
    returnPressed = Qt.pyqtSignal()

    DefaultIntDigitCount = 6
    DefaultDecDigitCount = 2

    def __init__(self, parent=None):
        """__init__(self, parent = None) -> QWheelEdit

        Constructor

        @param[in] parent (QWidget) the parent widget (optional, default is None
                          meaning there is no parent
        """
        Qt.QFrame.__init__(self, parent)
        Qt.QGridLayout(self)
        self._roundFunc = None
        self._previous_value = 0
        self._value = 0
        self._value_str = '0'
        self._minValue = numpy.finfo('d').min  # -inf
        self._maxValue = numpy.finfo('d').max  # inf
        self._editor = None
        self._editing = False
        self._showArrowButtons = True
        self._setDigits(QWheelEdit.DefaultIntDigitCount,
                        QWheelEdit.DefaultDecDigitCount)
        self._setValue(0)

        self._build()

    def _getMinPossibleValue(self):
        """_getMinPossibleValue(self) -> None

        Determines which is the minimum possible value that can be represented
        with the current total number of digits.

        @return (float) the minimum possible value
        """
        decmax = 0
        for i in xrange(self.getDecDigitCount()):
            decmax += 9 * math.pow(10, -(i + 1))
        return -math.pow(10.0, self.getIntDigitCount()) + 1 - decmax

    def _getMaxPossibleValue(self):
        """_getMaxPossibleValue(self) -> None

        Determines which is the maximum possible value that can be represented
        with the current total number of digits.

        @return (float) the maximum possible value
        """
        decmax = 0
        for i in xrange(self.getDecDigitCount()):
            decmax += 9 * math.pow(10, -(i + 1))
        return math.pow(10.0, self.getIntDigitCount()) - 1 + decmax

    def _build(self):
        """_build(self) -> None

        Builds this widget sub-items"""

        l = self.layout()
        l.setSpacing(0)
        l.setMargin(0)

        id = self.getIntDigitCount()
        dd = self.getDecDigitCount()
        digits = self.getDigitCount()

        self._upButtons = Qt.QButtonGroup()
        self._downButtons = Qt.QButtonGroup()
        self._digitLabels = []
        for l in self._digitLabels:
            l.setAlignment(Qt.AlignCenter)

        showDot = self.getDecDigitCount() > 0

        signLabel = _DigitLabel('+')
        signLabel.setFocusPolicy(Qt.Qt.NoFocus)
        signLabel.setAlignment(Qt.Qt.AlignRight | Qt.Qt.AlignVCenter)
        self._digitLabels.append(signLabel)
        l.addWidget(signLabel, 1, 0)
        l.setRowMinimumHeight(1, signLabel.minimumSizeHint().height())
        l.setColumnMinimumWidth(0, _ArrowButton.ButtonSize)
        l.setColumnStretch(0, 1)

        for i in xrange(id):
            col = i + 1
            d = _DigitLabel('0')
            up = _UpArrowButton(id - i - 1)
            down = _DownArrowButton(id - i - 1)
            d.setButtons(up, down)
            up.setFocusProxy(d)
            down.setFocusProxy(d)
            self._upButtons.addButton(up, col)
            self._downButtons.addButton(down, col)
            self._digitLabels.append(d)
            if self.getShowArrowButtons():
                l.addWidget(up, 0, col)
                l.addWidget(down, 2, col)
            l.addWidget(d, 1, col)

        if showDot:
            dotLabel = _DigitLabel('.')
            dotLabel.setFocusPolicy(Qt.Qt.NoFocus)
            dotLabel.setAlignment(Qt.Qt.AlignCenter)
            self._digitLabels.append(dotLabel)
            l.addWidget(dotLabel, 1, id + 1)

        for i in xrange(id, digits):
            col = i + 1
            if showDot:
                col += 1
            d = _DigitLabel('0')
            up = _UpArrowButton(id - i - 1)
            down = _DownArrowButton(id - i - 1)
            d.setButtons(up, down)
            up.setFocusProxy(d)
            down.setFocusProxy(d)
            self._upButtons.addButton(up, col)
            self._downButtons.addButton(down, col)
            self._digitLabels.append(d)
            if self.getShowArrowButtons():
                l.addWidget(up, 0, col)
                l.addWidget(down, 2, col)
            l.addWidget(d, 1, col)

        self._upButtons.buttonClicked.connect(self.buttonPressed)
        self._downButtons.buttonClicked.connect(self.buttonPressed)

        ed = _NumericEditor(self)
        ed.returnPressed.connect(self.editingFinished)
        ed.lostFocus.connect(ed.hide)
        rect = Qt.QRect(l.cellRect(1, 0).topLeft(), l.cellRect(1,
                                                               l.columnCount() - 1).bottomRight())
        ed.setGeometry(rect)
        ed.setAlignment(Qt.Qt.AlignRight)
        ed.validator().setRange(self.getMinValue(), self.getMaxValue(),
                                self.getDecDigitCount())
        ed.setVisible(False)
        self._editor = ed
        self.clearWarning()

    def _clear(self):
        """_clear(self) -> None

        Clears this widget sub-items"""

        self._upButtons.buttonClicked.disconnect(self.buttonPressed)
        self._downButtons.buttonClicked.disconnect(self.buttonPressed)

        for b in self._upButtons.buttons():
            self._upButtons.removeButton(b)
            b.setParent(None)
            b.destroy()
        for b in self._downButtons.buttons():
            self._downButtons.removeButton(b)
            b.setParent(None)
            b.destroy()
        for label in self._digitLabels:
            label.setParent(None)
            label.destroy()
        self._digitLabels = None
        self._upButtons.setParent(None)
        self._downButtons.setParent(None)
        self._upButtons = None
        self._downButtons = None
        self._editor.setParent(None)
        self._editor.destroy()
        self._editor = None

    def _setMinMax(self, min, max):
        """_setMinMax(self, min, max) -> None

        Sets the minimum and maximum values for this widget

        @param[in] min(float) minimum value
        @param[in] max(float) maximum value
        """
        self._minValue = min
        self._maxValue = max

    def _setDigits(self, int_nb=None, dec_nb=None):
        """_setDigits(self, int_nb=None, dec_nb=None) -> None

        Sets the number of digits that this widget shows. It will ensure that
        the current value can be displayed (i.e., int_nb will be forced to be
        large enough for displaying the integer representation of the value)

        @param[in] int_nb (int) number of integer digits (optional, default is
                   None, meaning use the existing number of integer digits
        @param[in] dec_nb (int) number of decimal digits (optional, default is
                   None, meaning use the existing number of decimal digits
        """

        if not int_nb is None:
            self._intDigitCount = int_nb
        if not dec_nb is None:
            self._decDigitCount = dec_nb

        # make sure that the current value can be displayed
        self._intDigitCount = max(self._intDigitCount, len("%d" % self._value))

        self._digitCount = self._intDigitCount + self._decDigitCount
        total_chars = self._digitCount
        total_chars += 1  # for sign
        if self._decDigitCount > 0:
            total_chars += 1  # for dot
        self._valueFormat = '%%+0%d.%df' % (total_chars, self._decDigitCount)
        self._setMinMax(self._getMinPossibleValue(),
                        self._getMaxPossibleValue())
        # we call setValue to update the self._value_str
        self._setValue(self.getValue())

    def _buildValueStr(self, v):
        """_buildValueStr(self, v) -> str

        Builds a string representation of the given value according to the
        number of digits this widget can currently represent

        @param[in] v (float) the value to be translated

        @return (str) a proper string representation of the given value
        """
        if v is None:
            ret = (self._valueFormat % 0).replace('0', '-')
        else:
            ret = self._valueFormat % v
            if ret.endswith('nan'):
                ret = ret.replace('0', ' ')
        self._value_str = ret
        return ret

    def _updateDigits(self):
        """_updateDigits(self) -> None

        Updates this widget contents (sub-widgets) according to the current
        widget configuration
        """
        self._clear()
        self._build()
        self._updateValue(False, False)

    def _updateValue(self, trigValueChanged=True, trigValueEdited=True):
        """_updateValue(self, trigValueChanged=True, trigValueEdited=True) -> None

        Updates this widget displayed value and potentially send signal(s)

        @param[in] trigValueChanged (bool) (optional, default is True) wheather or
                   not to send signal 'valueChanged(double)'
        @param[in] trigValueEdited (bool) (optional, default is True) wheather or
                   not to send signal 'valueEdit(double)'
        """
        v, v_str = self.getValue(), self.getValueStr()

        if len(v_str) > len(self._digitLabels):
            # do auto adjust
            if '.' in v_str:
                dc = map(len, v_str.split('.'))
            else:
                dc = len(v_str), 0
            self._setDigits(*dc)
            v_str = self._buildValueStr(v)
            self._clear()
            self._build()

        for i, c in enumerate(v_str):
            self._digitLabels[i].setText(c)

        if trigValueChanged:
            self.numberChanged.emit(v)

        if trigValueEdited:
            self.numberEdited.emit(v)

    def setRoundFunc(self, roundFunc):
        """setRoundFunc(self, roundFunc) -> None

        Sets the rounding function to use when calling _setValue(). This allows you to
        filter invalid user input

        @param[in] roundFunc (callable) the rounding function to use
        """
        self._roundFunc = roundFunc

    def getPreviousValue(self):
        """getPreviousValue(self) -> float

        Gives the previous value of this widget

        @return (float) the previous value of this widget
        """
        return self._previous_value

    def _setValue(self, v):
        """_setValue(self, v) -> None

        Sets value of this widget. If the given value exceeds any limit, the
        value is NOT set.
        """
        if self._roundFunc:
            v = self._roundFunc(v)
        if v > self._maxValue or v < self._minValue:
            return
        self._previous_value = self._value
        self._value = v
        self._buildValueStr(v)

    def setWarning(self, msg):
        """setWarning(self, msg) -> None

        Activates the warning style for this widget. This means a violet border
        and a tooltip with the given message.

        @param[in] msg (str) the message to be displayed as tooltip
        """
        self.setStyleSheet(
            'QWheelEdit {border: 1px solid; border-radius: 4px; border-color: violet}')
        self.setToolTip(msg)

    def clearWarning(self):
        """clearWarning(self) -> None

        Clears the warning style. If not in warning mode, nothing is done.
        """
        self.setStyleSheet(
            'QWheelEdit {border: 1px solid; border-radius: 4px; border-color: rgba(0,0,0,0)}')
        self.setToolTip('')

    def buttonPressed(self, b):
        """buttonPressed(self, b) -> None

        Slot executed when an arrow button is pressed from the button group

        @param[in] b (_ArrowButton) the button which was pressed
        """
        self._setValue(self.getValue() + b._inc)
        self._updateValue()

    def setDigitCount(self, int_nb, dec_nb):
        """setDigitCount(self, int_nb, dec_nb) -> None

        Updates the displayed digits.

        @param[in] int_nb(int) number of integer digits
        @param[in] dec_nb(int) number of decimal digits
        """
        self._setDigits(int_nb=int_nb, dec_nb=dec_nb)
        self._updateDigits()

    def getDigitCount(self):
        """getDigitCount(self) -> int

        Gets the total number of digits this widget displays

        @return (int) the total number of digits this widget displays
        """
        return self._digitCount

    def getIntDigitCount(self):
        """getIntDigitCount(self) -> int

        Gets the number of integer digits this widget displays

        @return (int) the number of integer digits this widget displays
        """
        return self._intDigitCount

    def setIntDigitCount(self, n):
        """setIntDigitCount(self, n) -> None

        Sets the number of integer digits this widget displays

        @param[in] n (int) the number of integer digits to display
        """
        self._setDigits(int_nb=n)
        self._updateDigits()

    def resetIntDigitCount(self):
        """resetIntDigitCount(self) -> None

        Resets the number of integer digits this widget displays to DefaultIntDigitCount
        """
        self.setIntDigitCount(QWheelEdit.DefaultIntDigitCount)

    def getDecDigitCount(self):
        """getDecDigitCount(self) -> int

        Gets the number of decimal digits this widget displays

        @return (int) the number of decimal digits this widget displays
        """
        return self._decDigitCount

    def setDecDigitCount(self, n):
        """setDecDigitCount(self, n) -> None

        Sets the number of decimal digits this widget displays

        @param[in] n (int) the number of decimal digits to display
        """

        self._setDigits(dec_nb=n)
        self._updateDigits()

    def resetDecDigitCount(self):
        """resetDecDigitCount(self) -> None

        Resets the number of decimal digits this widget displays to DefaultDecDigitCount
        """
        self.setDecDigitCount(QWheelEdit.DefaultDecDigitCount)

    def setValue(self, v):
        """setValue(self, v) -> None

        Sets the value of this widget.
        Send a 'valueChanged(double)' Qt signal

        @param[in] v (float/Quantity) the value to be set
        """
        if isinstance(v, Q_):
            v = v.magnitude
        self._setValue(v)
        self._updateValue(trigValueEdited=False)

    def resetValue(self):
        """resetValue(self) -> None

        Resets the value of this widget to 0.0
        """
        self.setValue(0.0)

    def getValue(self):
        """getValue(self) -> float

        Gets the current value of this widget

        @return (float) the value currently displayed by the widget
        """
        return self._value

    def getValueStr(self):
        """getValueStr(self) -> str

        Gets the current value string of this widget

        @return (str) the value currently displayed by the widget
        """
        return self._value_str

    def getMinValue(self):
        """getMinValue(self) -> float

        Gets the minimum allowed value

        @return (float) the minimum allowed value
        """
        return self._minValue

    def setMinValue(self, v):
        """setMinValue(self, v) -> None

        Sets the minimum allowed value for the widget

        @param[in] v (float) the new minimum allowed value
        """
        self._minValue = v
        w = self.getEditWidget()
        if not w is None:
            w.validator().setRange(self._minValue, self._maxValue, self._decDigitCount)

    def resetMinValue(self):
        """resetMinValue(self) -> None

        Resets the minimum allowed value to the minimum possible according to
        the current total number of digits
        """
        self.setMinValue(self._getMinPossibleValue())

    def getMaxValue(self):
        """getMaxValue(self) -> float

        Gets the maximum allowed value

        @return (float) the maximum allowed value
        """
        return self._maxValue

    def setMaxValue(self, v):
        """setMaxValue(self, v) -> None

        Sets the maximum allowed value for the widget

        @param[in] v (float) the new maximum allowed value
        """
        self._maxValue = v
        w = self.getEditWidget()
        if not w is None:
            w.validator().setRange(self._minValue, self._maxValue, self._decDigitCount)

    def resetMaxValue(self):
        """resetMaxValue(self) -> None

        Resets the maximum allowed value to the maximum possible according to
        the current total number of digits
        """
        self.setMaxValue(self._getMaxPossibleValue())

    def getAutoRepeat(self):
        return self._upButtons.buttons()[0].autoRepeat()

    def setAutoRepeat(self, v):
        for b in self._upButtons.buttons():
            b.setAutoRepeat(v)
        for b in self._downButtons.buttons():
            b.setAutoRepeat(v)

    def resetAutoRepeat(self):
        self.setAutoRepeat(False)

    def getAutoRepeatDelay(self):
        return self._upButtons.buttons()[0].autoRepeatDelay()

    def setAutoRepeatDelay(self, milisecs):
        for b in self._upButtons.buttons():
            b.setAutoRepeatDelay(milisecs)
        for b in self._downButtons.buttons():
            b.setAutoRepeatDelay(milisecs)

    def getAutoRepeatInterval(self):
        return self._upButtons.buttons()[0].autoRepeatInterval()

    def setAutoRepeatInterval(self, milisecs):
        for b in self._upButtons.buttons():
            b.setAutoRepeatInterval(milisecs)
        for b in self._downButtons.buttons():
            b.setAutoRepeatInterval(milisecs)

    def getEditWidget(self):
        """getEditWidget(self) -> QWidget

        Gets the widget object used when the user manually sets the value

        @return (QWidget) the widget used for editing
        """
        return self._editor

    def editingFinished(self):
        """editingFinished(self) -> None

        Slot called when the user finishes editing
        """
        ed = self.getEditWidget()
        v = float(ed.text())
        self._setValue(v)
        self._updateValue()

    def showEditWidget(self):
        """showEditWidget(self) -> None

        Forces the edition widget to be displayed
        """
        ed, l = self.getEditWidget(), self.layout()
        rect = Qt.QRect(l.cellRect(1, 0).topLeft(), l.cellRect(1,
                                                               l.columnCount() - 1).bottomRight())
        ed.setGeometry(rect)
        ed.setAlignment(Qt.Qt.AlignRight)
        ed.setMaxLength(self.getDigitCount() + 2)
        ed.setText(self.getValueStr())
        ed.selectAll()
        ed.setFocus()
        ed.setVisible(True)

    def hideEditWidget(self):
        """hideEditWidget(self) -> None

        Forces the edition widget to be hidden
        """
        ed = self.getEditWidget()
        ed.setVisible(False)
        self.setFocus()

    def wheelEvent(self, evt):
        numDegrees = evt.delta() / 8
        numSteps = numDegrees / 15
        #w = Qt.QApplication.focusWidget()
        w = self.focusWidget()
        if not isinstance(w, _DigitLabel):
            return Qt.QFrame.wheelEvent(self, evt)

        if numSteps > 0:
            inc = w._upButton._inc
        else:
            inc = -w._downButton._inc
        numSteps *= inc
        self.setValue(self.getValue() + numSteps)
        evt.accept()

    def mouseDoubleClickEvent(self, mouse_event):
        """mouseDoubleClickEvent(self, mouse_event)

        Executed when user presses double click. This widget shows the edition
        widget when this happens
        """
        self.showEditWidget()
        self._editing = True

    def keyPressEvent(self, key_event):
        """keyPressEvent(self, key_event) -> None

        Exectuted when the user presses a key.
        F2 enters/leaves edition mode. ESC leaves edition mode
        """
        k = key_event.key()
        if k == Qt.Qt.Key_F2:
            if self._editing:
                self.hideEditWidget()
                self._editing = False
            else:
                self.showEditWidget()
                self._editing = True
            return
        elif k == Qt.Qt.Key_Escape:
            if self._editing:
                self.hideEditWidget()
                self._editing = False
                return
        elif k in (Qt.Qt.Key_Return, Qt.Qt.Key_Enter):
            if self._editing:
                self.hideEditWidget()
                self._editing = False
            else:
                self.returnPressed.emit()

        # TODO Decide when to emit editingFinished for completeness
        Qt.QWidget.keyPressEvent(self, key_event)

    def getShowArrowButtons(self):
        return self._showArrowButtons

    def setShowArrowButtons(self, yesno):
        self._showArrowButtons = yesno
        self._updateDigits()

    def resetShowArrowButtons(self):
        self.setShowArrowButtons(True)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    showArrowButtons = Qt.pyqtProperty("bool", getShowArrowButtons,
                                       setShowArrowButtons,
                                       resetShowArrowButtons)

    integerDigits = Qt.pyqtProperty("int", getIntDigitCount,
                                    setIntDigitCount, resetIntDigitCount)

    decimalDigits = Qt.pyqtProperty("int", getDecDigitCount,
                                    setDecDigitCount, resetDecDigitCount)

    value = Qt.pyqtProperty("double", getValue,
                            setValue, resetValue)

    minValue = Qt.pyqtProperty("double", getMinValue,
                               setMinValue, resetMinValue)

    maxValue = Qt.pyqtProperty("double", getMaxValue,
                               setMaxValue, resetMaxValue)

    autoRepeat = Qt.pyqtProperty("bool", getAutoRepeat,
                                 setAutoRepeat, resetAutoRepeat)

    autoRepeatDelay = Qt.pyqtProperty("int", getAutoRepeatDelay,
                                      setAutoRepeatDelay)

    autoRepeatInterval = Qt.pyqtProperty("int", getAutoRepeatInterval,
                                         setAutoRepeatInterval)


def main():
    global arrowWidget

    def resetAll():
        arrowWidget.resetIntDigitCount()
        arrowWidget.resetDecDigitCount()
        arrowWidget.resetMinValue()
        arrowWidget.resetMaxValue()
        arrowWidget.resetValue()

    def setNAN():
        arrowWidget.setValue(float('nan'))

    def setNone():
        arrowWidget.setValue(None)
    a = Qt.QApplication([])
    panel = Qt.QWidget()
    l = Qt.QFormLayout(panel)
    button_layout = Qt.QVBoxLayout()
    arrowWidget = QWheelEdit(panel)
    isb, dsb = Qt.QSpinBox(panel), Qt.QSpinBox(panel)
    minv, maxv = Qt.QDoubleSpinBox(panel), Qt.QDoubleSpinBox(panel)
    resetbutton = Qt.QPushButton("Reset", panel)
    resetbutton.setDefault(True)
    nanbutton = Qt.QPushButton("Set NAN", panel)
    nonebutton = Qt.QPushButton("Set None", panel)
    showarrowbutton = Qt.QCheckBox("", panel)

    l.addRow("Value", arrowWidget)
    l.addRow("Integer digits:", isb)
    l.addRow("Decimal digits:", dsb)
    l.addRow("Minimum value:", minv)
    l.addRow("Maximum value:", maxv)
    l.addRow("Show arrows:", showarrowbutton)
    l.addRow(button_layout)
    button_layout.addWidget(nanbutton)
    button_layout.addWidget(nonebutton)
    button_layout.addWidget(resetbutton)
    isb.setValue(arrowWidget.getIntDigitCount())
    dsb.setValue(arrowWidget.getDecDigitCount())
    minv.setRange(numpy.finfo('d').min, numpy.finfo('d').max)
    maxv.setRange(numpy.finfo('d').min, numpy.finfo('d').max)
    minv.setValue(arrowWidget.getMinValue())
    maxv.setValue(arrowWidget.getMaxValue())
    showarrowbutton.setChecked(arrowWidget.getShowArrowButtons())
    isb.valueChanged.connect(arrowWidget.setIntDigitCount)
    dsb.valueChanged.connect(arrowWidget.setDecDigitCount)
    minv.valueChanged.connect(arrowWidget.setMinValue)
    showarrowbutton.stateChanged.connect(arrowWidget.setShowArrowButtons)
    nanbutton.clicked.connect(setNAN)
    nonebutton.clicked.connect(setNone)
    resetbutton.clicked.connect(resetAll)
    panel.setVisible(True)
    a.exec_()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Adapted from http://code.activestate.com/recipes/267662/"""

__docformat__ = "restructuredtext"

import cStringIO
import operator
import re
import math


def indent(rows, hasHeader=False, headerChar='-', delim=' | ', justify='left',
           separateRows=False, prefix='', postfix='', wrapfunc=lambda x: x):
    """Indents a table by column.
       - rows: A sequence of sequences of items, one sequence per row.
       - hasHeader: True if the first row consists of the columns' names.
       - headerChar: Character to be used for the row separator line
         (if hasHeader==True or separateRows==True).
       - delim: The column delimiter.
       - justify: Determines how are data justified in their column.
         Valid values are 'left','right' and 'center'.
       - separateRows: True if rows are to be separated by a line
         of 'headerChar's.
       - prefix: A string prepended to each printed row.
       - postfix: A string appended to each printed row.
       - wrapfunc: A function f(text) for wrapping text; each element in
         the table is first wrapped by this function.

        Returns a list of strings. One for each row of the table
    """

    # closure for breaking logical rows to physical, using wrapfunc
    def rowWrapper(row):
        newRows = [wrapfunc(item).split('\n') for item in row]
        return [[substr or '' for substr in item] for item in map(None, *newRows)]

    # break each logical row into one or more physical ones
    logicalRows = [rowWrapper(row) for row in rows]
    # columns of physical rows

    columns = map(None, *reduce(operator.add, logicalRows))

    # get the maximum of each column by the string length of its items
    maxWidths = [max([len(str(item)) for item in column])
                 for column in columns]
    if separateRows or hasHeader:
        rowSeparator = headerChar * (len(prefix) + len(postfix) + sum(maxWidths) +
                                     len(delim) * (len(maxWidths) - 1))
    else:
        rowSeparator = "<ERR>"

    # select the appropriate justify method
    justify = {'center': str.center, 'right': str.rjust,
               'left': str.ljust}[justify.lower()]

    output = []
    if separateRows:
        output.append(rowSeparator)
    for physicalRows in logicalRows:
        for row in physicalRows:
            line = prefix
            line += delim.join([justify(str(item), width)
                                for (item, width) in zip(row, maxWidths)])
            line += postfix
            output.append(line)
        if separateRows or hasHeader:
            output.append(rowSeparator)
            hasHeader = False
    return output

# written by Mike Brown
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/148061


def wrap_onspace(text, width):
    """
    A word-wrap function that preserves existing line breaks
    and most spaces in the text. Expects that existing line
    breaks are posix newlines (\n).
    """
    return reduce(lambda line, word, width=width: '%s%s%s' %
                  (line,
                   ' \n'[(len(line[line.rfind('\n') + 1:])
                          + len(word.split('\n', 1)[0]
                                ) >= width)],
                   word),
                  text.split(' ')
                  )


def wrap_onspace_strict(text, width):
    """Similar to wrap_onspace, but enforces the width constraint:
       words longer than width are split."""
    wordRegex = re.compile(r'\S{' + str(width) + r',}')
    return wrap_onspace(wordRegex.sub(lambda m: wrap_always(m.group(), width), text), width)

import math


def wrap_always(text, width):
    """A simple word-wrap function that wraps text on exactly width characters.
       It doesn't split the text in words."""
    return '\n'.join([text[width * i:width * (i + 1)]
                      for i in xrange(int(math.ceil(1. * len(text) / width)))])

if __name__ == '__main__':
    labels = ('First Name', 'Last Name', 'Age', 'Position')
    data = \
        '''John,Smith,24,Software Engineer
       Mary,Brohowski,23,Sales Manager
       Aristidis,Papageorgopoulos,28,Senior Reseacher'''
    rows = [row.strip().split(',') for row in data.splitlines()]

    print 'Without wrapping function\n'
    for l in indent([labels] + rows, hasHeader=True):
        print l

    # test indent with different wrapping functions
    width = 10
    for wrapper in (wrap_always, wrap_onspace, wrap_onspace_strict):
        print 'Wrapping function: %s(x,width=%d)\n' % (wrapper.__name__, width)
        o = indent([labels] + rows, headerChar='=', hasHeader=True, separateRows=False,
                   prefix='|', postfix='|', delim=' ',
                   wrapfunc=lambda x: wrapper(x, width))
        for l in o:
            print l

    # output:
    #
    # Without wrapping function
    #
    # First Name | Last Name        | Age | Position
    #-------------------------------------------------------
    # John       | Smith            | 24  | Software Engineer
    # Mary       | Brohowski        | 23  | Sales Manager
    # Aristidis  | Papageorgopoulos | 28  | Senior Reseacher
    #
    # Wrapping function: wrap_always(x,width=10)
    #
    #----------------------------------------------
    #| First Name | Last Name  | Age | Position   |
    #----------------------------------------------
    #| John       | Smith      | 24  | Software E |
    #|            |            |     | ngineer    |
    #----------------------------------------------
    #| Mary       | Brohowski  | 23  | Sales Mana |
    #|            |            |     | ger        |
    #----------------------------------------------
    #| Aristidis  | Papageorgo | 28  | Senior Res |
    #|            | poulos     |     | eacher     |
    #----------------------------------------------
    #
    # Wrapping function: wrap_onspace(x,width=10)
    #
    #---------------------------------------------------
    #| First Name | Last Name        | Age | Position  |
    #---------------------------------------------------
    #| John       | Smith            | 24  | Software  |
    #|            |                  |     | Engineer  |
    #---------------------------------------------------
    #| Mary       | Brohowski        | 23  | Sales     |
    #|            |                  |     | Manager   |
    #---------------------------------------------------
    #| Aristidis  | Papageorgopoulos | 28  | Senior    |
    #|            |                  |     | Reseacher |
    #---------------------------------------------------
    #
    # Wrapping function: wrap_onspace_strict(x,width=10)
    #
    #---------------------------------------------
    #| First Name | Last Name  | Age | Position  |
    #---------------------------------------------
    #| John       | Smith      | 24  | Software  |
    #|            |            |     | Engineer  |
    #---------------------------------------------
    #| Mary       | Brohowski  | 23  | Sales     |
    #|            |            |     | Manager   |
    #---------------------------------------------
    #| Aristidis  | Papageorgo | 28  | Senior    |
    #|            | poulos     |     | Reseacher |
    #---------------------------------------------



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a base widget that can be used to display a taurus
model in a table widget"""

# todo: tango-centric!!!

__all__ = ["TaurusDbTableWidget"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.core.taurusbasetypes import TaurusElementType
from taurus.qt.qtcore.model import *
from taurus.core.taurusauthority import TaurusAuthority
from taurus.qt.qtgui.icon import getElementTypeIcon, getElementTypeIconName
from taurustable import TaurusBaseTableWidget


class TaurusDbTableWidget(TaurusBaseTableWidget):
    """A class:`taurus.qt.qtgui.tree.TaurusBaseTableWidget` that connects to a
    :class:`taurus.core.taurusauthority.TaurusAuthority` model. It can show
    the list of database elements in two different perspectives:

    - device : a device list based perspective
    - server : a server list based perspective

    Filters can be inserted into this widget to restrict the items that are
    seen.
    """

    KnownPerspectives = {
        TaurusElementType.Device: {
            "label": "Devices",
            "icon": getElementTypeIconName(TaurusElementType.Device),
            "tooltip": "View by device",
            "model": [TaurusDbDeviceProxyModel, TaurusDbBaseModel, ],
        },
        TaurusElementType.Server: {
            "label": "Servers",
            "icon": getElementTypeIconName(TaurusElementType.Server),
            "tooltip": "View by server",
            "model": [TaurusDbServerProxyModel, TaurusDbPlainServerModel, ],
        },
    }

    DftPerspective = TaurusElementType.Device

    def getModelClass(self):
        return TaurusAuthority

    def sizeHint(self):
        return Qt.QSize(1024, 512)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseTableWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.table'
        ret['group'] = 'Taurus Views'
        ret['icon'] = "designer:table.png"
        return ret


def main_TaurusDbTableWidget(host, perspective=TaurusElementType.Device):
    w = TaurusDbTableWidget(perspective=perspective)
    w.setWindowIcon(getElementTypeIcon(perspective))
    w.setWindowTitle("A Taurus Table Example")
    w.setModel(host)
    w.show()
    return w


def demo():
    """Table panels"""
    import taurus
    db = taurus.Authority()
    host = db.getNormalName()
    w = main_TaurusDbTableWidget(host, TaurusElementType.Device)

    return w


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application(app_name="DB model demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")
    w = demo()
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



'''
This code checks if a known bug in some PyQwt versions affects your installation.
Simply run this script and see the diagnosis in the std output.

If you are affected, the solution is to install/compile a newer version of PyQwt/SIP

Notably, the python-qwt5-qt4 package originally shipped with
Ubuntu 10.10 and Debian 6 is affected

In some systems, the bug produces a segfault while in some others it
produces an Assertion error similar to the following:
python: /build/buildd/sip4-qt3-4.10.5/siplib/siplib.c:2600: sip_api_parse_result: Assertion `assign_helper != ((void *)0)' failed.

See also:
https://bugs.launchpad.net/ubuntu/+source/pyqwt5/+bug/672509
http://www.esrf.eu/mail_archives/tango/archive/msg04025.html
'''

from PyQt4 import Qt, Qwt5


class MyScaleDrawSafe(Qwt5.QwtScaleDraw):

    def __init__(self):
        Qwt5.QwtScaleDraw.__init__(self)


class MyScaleDrawDanger(Qwt5.QwtScaleDraw):

    def __init__(self):
        Qwt5.QwtScaleDraw.__init__(self)

    def label(self, val):
        return Qwt5.QwtScaleDraw.label(self, val)


class MyPlot(Qwt5.QwtPlot):

    def __init__(self, parent=None):
        Qwt5.QwtPlot.__init__(self, parent)
        self.setAxisScaleDraw(Qwt5.QwtPlot.xBottom, MyScaleDrawSafe())
        print "Replotting with MyScaleDrawSafe:..."
        self.replot()
        print "ok"
        self.setAxisScaleDraw(Qwt5.QwtPlot.xBottom, MyScaleDrawDanger())
        print "Replotting with MyScaleDrawDanger (if it crashes now you are affected by the bug) :..."
        self.replot()
        print "SAFE!!!"
        print "if this is printed, the sip/PyQwt bug does not affect you"

app = Qt.QApplication([])
p = MyPlot()
p.show()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a panel to display taurus messages"""

__all__ = ["TicketReportHandler"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from basicreport import SendMailDialog, SMTPReportHandler


class SendTicketDialog(SendMailDialog):

    def __init__(self, parent=None):
        SendMailDialog.__init__(self, parent=parent)
        self.ui.editTo.setParent(None)
        self.ui.editTo = Qt.QComboBox(self)
        self.ui.editTo.setEditable(True)
        self.ui.editTo.addItems(["controls", "mis", "electronics",
                                 "systems"])
        self.ui.editTo.setCurrentIndex(0)
        self.ui.mainLayout.addWidget(self.ui.editTo, 1, 1, 1, 1)

    def getTo(self):
        return str(self.ui.editTo.currentText() + "@rt.cells.es")


class TicketReportHandler(SMTPReportHandler):
    """Report a message by sending an ALBA ticket"""

    Label = "Send ticket"

    def getDialogClass(self):
        return SendTicketDialog


def main():
    app = Qt.QApplication([])
    w = SendTicketDialog()
    w.exec_()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""
configuration file for an example of how to construct a GUI based on TaurusGUI

This configuration file determines the default, permanent, pre-defined
contents of the GUI. While the user may add/remove more elements at run
time and those customizations will also be stored, this file defines what a
user will find when launching the GUI for the first time.
"""

#==============================================================================
# Import section. You probably want to keep this line. Don't edit this block
# unless you know what you are doing
from taurus.qt.qtgui.taurusgui.utils import PanelDescription, ExternalApp, Qt_Qt
# (end of import section)
#==============================================================================


#=========================================================================
# General info.
#=========================================================================
GUI_NAME = 'MacroGUI'
ORGANIZATION = 'Taurus'
# CUSTOM_LOGO = <full path to GUI-specific logo>

#=========================================================================
# If you want to have a main synoptic panel, set the SYNOPTIC variable
# to the file name of a jdraw file. If a relative path is given, the directory
# containing this configuration file will be used as root
# (comment out or make SYNOPTIC=None to skip creating a synoptic panel)
#=========================================================================
# SYNOPTIC =

#=========================================================================
# Set INSTRUMENTS_FROM_POOL to True for enabling auto-creation of
# instrument panels based on the Pool Instrument info
#=========================================================================
#INSTRUMENTS_FROM_POOL = False

#=========================================================================
# Define panels to be shown.
# To define a panel, instantiate a PanelDescription object (see documentation
# for the gblgui_utils module)
#=========================================================================

# None defined

#=========================================================================
# Define which External Applications are to be inserted.
# To define an external application, instantiate an ExternalApp object
# See TaurusMainWindow.addExternalAppLauncher for valid values of ExternalApp
#=========================================================================
#xterm = ExternalApp(cmdargs=['xterm','spock'], text="Spock", icon='utilities-terminal')
#hdfview = ExternalApp(["hdfview"])
pymca = ExternalApp(['pymca'])

#=========================================================================
# Macro execution configuration
# Comment out or make MACRO_SERVER=None or set MACRO_PANELS=False to skip
# creating a macro execution infrastructure.
# Give empty strings if you want to select the values manually in the GUI
#=========================================================================
MACROSERVER_NAME = ''
DOOR_NAME = ''
MACROEDITORS_PATH = ''

#=========================================================================
# Define one or more embedded consoles in the GUI.
# Possible items for console are 'ipython', 'tango', 'spock'
#=========================================================================
CONSOLE = ['tango']

#=========================================================================
# Monitor widget
#=========================================================================
# MONITOR =



# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Taurus
##
## http://taurus-scada.org
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module exposes QtCore module"""

from taurus.external.qt import getQtName

__backend = getQtName()


def __to_qvariant_1(pyobj=None):
    """Properly converts a python object into a proper QVariant according to
    the PySide or PyQt API version in use

    :param pyobj: object to be converted
    :return: A proper QVariant"""
    from PyQt4.QtCore import QVariant
    if pyobj is None:
        return QVariant()  # PyQt 4.4 doesn't accept QVariant(None)
    return QVariant(pyobj)


def __from_qvariant_1(qobj=None, convfunc=None):
    """Properly converts a QVariant/QVariant equivalent to a python object
    according to the PySide or PyQt API version in use

    :param qobj: object to be converted
    :param convfunc:
        conversion function. Not used if QVariant is not available.
        If QVariant is available:  [default: None, meaning use
        qobj.toPyObject()]. Can be a function like str, int, bool, float or
        a string containing the conversion method (ex.: 'toByteArray') will
        call qobj.toByteArray()
    :return: A proper python object"""
    if convfunc is None:
        return qobj.toPyObject()
    elif callable(convfunc):
        if convfunc in (unicode, str):
            return convfunc(qobj.toString())
        elif convfunc is bool:
            return qobj.toBool()
        elif convfunc is int:
            return qobj.toInt()[0]
        elif convfunc is float:
            return qobj.toDouble()[0]
    elif isinstance(convfunc, (str, unicode)):
        return getattr(qobj, convfunc)()


def __QVariant_2(pyobj=None):
    return pyobj


def __to_qvariant_2(pyobj=None):
    """Properly converts a python object into a proper QVariant according to
    the PySide or PyQt4 API version in use

    :param pyobj: object to be converted
    :return: A proper QVariant"""
    return pyobj


def __from_qvariant_2(qobj=None, convfunc=None):
    """Properly converts a QVariant/QVariant equivalent to a python object
    according to the PySide or PyQt4 API version in use

    :param qobj: object to be converted
    :param convfunc:
        conversion function. Not used if QVariant is not available.
        If QVariant is available:  [default: None, meaning use
        qobj.toPyObject()]. Can be a function like str, int, bool, float or
        a string containing the conversion method (ex.: 'toByteArray') will
        call qobj.toByteArray()
    :return: A proper python object"""
    return qobj


__QString_2 = str

__QStringList_2 = list


if __backend == 'PyQt4':
    from PyQt4 import QtCore as __QtCore

    # Alias PyQt-specific functions for PySide compatibility.
    if hasattr(__QtCore, "pyqtSignal"):
        Signal = __QtCore.pyqtSignal
    if hasattr(__QtCore, "pyqtSlot"):
        Slot = __QtCore.pyqtSlot
    else: #implement dummy pyqtSlot decorator for PyQt<4.6
        class DummyPyqtSlot(object):
            def __init__(self, *a, **kw):
                pass
            def __call__(self, f):
                return f
        Slot = pyqtSlot = DummyPyqtSlot
    if hasattr(__QtCore, "pyqtProperty"):
        Property = __QtCore.pyqtProperty

    try:
        __api_version__ = __QtCore.QT_VERSION_STR
    except AttributeError:
        pass

    try:
        import sip
        PYQT_QVARIANT_API_1 = sip.getapi('QVariant') < 2
        PYQT_QSTRING_API_1 = sip.getapi('QString') < 2
    except (ImportError, AttributeError):
        PYQT_QVARIANT_API_1 = True
        PYQT_QSTRING_API_1 = True

    if PYQT_QVARIANT_API_1:
        to_qvariant = __to_qvariant_1
        from_qvariant = __from_qvariant_1
    else:
        __QtCore.QVariant = QVariant = __QVariant_2
        to_qvariant = __to_qvariant_2
        from_qvariant = __from_qvariant_2

    if not PYQT_QSTRING_API_1:
        __QtCore.QString = QString = __QString_2
        __QtCore.QStringList = QStringList = __QStringList_2

    from PyQt4.QtCore import *

elif __backend == 'PyQt5':
    from PyQt5 import QtCore as __QtCore

    # Alias PyQt-specific functions for PySide compatibility.
    if hasattr(__QtCore, "pyqtSignal"):
        Signal = __QtCore.pyqtSignal
    if hasattr(__QtCore, "pyqtSlot"):
        Slot = __QtCore.pyqtSlot
    if hasattr(__QtCore, "pyqtProperty"):
        Property = __QtCore.pyqtProperty

    try:
        __api_version__ = __QtCore.QT_VERSION_STR
    except AttributeError:
        pass

    __QtCore.QVariant = QVariant = __QVariant_2
    to_qvariant = __to_qvariant_2
    from_qvariant = __from_qvariant_2
    __QtCore.QString = QString = __QString_2
    __QtCore.QStringList = QStringList = __QStringList_2

    from PyQt5.QtCore import *

elif __backend == 'PySide':
    from PySide import QtCore as __QtCore
    from PySide.QtCore import *
    __api_version__ = __QtCore.QT_VERSION_STR

    __QtCore.QVariant = QVariant = __QVariant_2
    to_qvariant = __to_qvariant_2
    from_qvariant = __from_qvariant_2
    __QtCore.QString = QString = __QString_2
    __QtCore.QStringList = QStringList = __QStringList_2

    #a dummy pyqtsignature decorator
    # CAUTION this totally nulifies the pupose of decorating with pyqtSignature
    # todo: do a proper implementation of pyqtsignature
    def pyqtSignature(f):
        return f

    # Alias PySide functions for PyQt compatibility.
    if hasattr(__QtCore, "Signal"):
        pyqtSignal = Signal
    if hasattr(__QtCore, "Slot"):
        pyqtSlot = Slot
    if hasattr(__QtCore, "Property"):
        pyqtProperty = Property

del getQtName



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.core.tango.test.test_tangovalidator..."""

# __all__ = []

__docformat__ = 'restructuredtext'

import numpy
import PyTango
from taurus.external import unittest
from taurus.external.pint import Quantity
import taurus
from taurus.core import DataType, DataFormat
from taurus.core.tango.tangoattribute import TangoAttrValue
from taurus.core.tango.test import TangoSchemeTestLauncher
from taurus.test import insertTest
from taurus.core.taurusbasetypes import AttrQuality
from taurus.core.tango.util.tango_taurus import unit_from_tango

_INT_IMG = numpy.arange(2 * 3, dtype='int16').reshape((2, 3))
_INT_SPE = _INT_IMG[1, :]
_FLOAT_IMG = _INT_IMG * .1
_FLOAT_SPE = _INT_SPE * .1
_BOOL_IMG = numpy.array([[True, False], [False, True]])
_BOOL_SPE = [True, False]
_STR = 'foo BAR |-+#@!?_[]{}'
_UCHAR_IMG = numpy.array([['a', 'b'], ['c', 'd']], dtype='S1')
_UCHAR_SPE = _UCHAR_IMG[1, :]
_UINT8_IMG = _UCHAR_IMG.view('uint8')
_UINT8_SPE = _UINT8_IMG[1, :]

# ==============================================================================
# Test writing fragment values
##
# DISABLED: these tests break test isolation (looks like Reset is not working
#           for configurations). Until we solve it, we disable all these tests
#
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar_nu',
#             cfg='range', value=[float('-inf'), float('inf')],
#             expected=[Quantity(float('-inf')), Quantity(float('inf'))]
#             )
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar_nu',
#             cfg='range', value=[Quantity(float('-inf')),
#                                 Quantity(float('inf'))],
#             expected=[Quantity(float('-inf')), Quantity(float('inf'))])
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar_nu',
#             cfg='range', value=[100, 300],
#             expected=[Quantity(100), Quantity(300)])
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar_nu',
#             cfg='range', value=[Quantity(100), Quantity(300)],
#             expected=[Quantity(100), Quantity(300)])
# @insertTest(helper_name='write_read_conf', attr_name='float_scalar',
#             cfg='range', value=[Quantity(-5, 'mm'), Quantity(5, 'mm')],
#             expected=[Quantity(-0.005, 'm'), Quantity(5, 'mm')])
# @insertTest(helper_name='write_read_conf', attr_name='short_spectrum',
#             cfg='label', value='Just a Test',
#             expected='Just a Test')
# @insertTest(helper_name='write_read_conf', attr_name='boolean_spectrum',
#             cfg='label', value='Just_a_Test',
#             expected='Just_a_Test')
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar',
#             cfg='warnings', value=[Quantity(-2, 'mm'), Quantity(2, 'mm')],
#             expected=[Quantity(-2, 'mm'), Quantity(0.002, 'm')])
# @insertTest(helper_name='write_read_conf', attr_name='short_image',
#             cfg='warnings', value=[Quantity(-2, 'mm'), Quantity(2, 'mm')],
#             expected=[Quantity(-0.002, 'm'), Quantity(2, 'mm')])
# @insertTest(helper_name='write_read_conf', attr_name='float_image',
#             cfg='warnings', value=[Quantity(-0.75, 'mm'),
#                                         Quantity(0.75, 'mm')],
#             expected=[Quantity(-0.00075, 'm'), Quantity(0.75, 'mm')])
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar_nu',
#             cfg='warnings', value=[100, 300],
#             expected=[Quantity(100), Quantity(300)])
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar',
#             cfg='alarms', value=[Quantity(-50, 'mm'), Quantity(50, 'mm')],
#             expected=[Quantity(-50, 'mm'), Quantity(50, 'mm')])
# @insertTest(helper_name='write_read_conf', attr_name='short_image',
#             cfg='alarms', value=[Quantity(-2, 'mm'), Quantity(2, 'mm')],
#             expected=[Quantity(-0.002, 'm'), Quantity(2, 'mm')])
# @insertTest(helper_name='write_read_conf', attr_name='float_image',
#             cfg='alarms', value=[Quantity(-0.75, 'mm'), Quantity(0.75, 'mm')],
#             expected=[Quantity(-0.00075, 'm'), Quantity(0.75, 'mm')])
# @insertTest(helper_name='write_read_conf', attr_name='short_scalar_nu',
#             cfg='alarms', value=[100, 300],
#             expected=[Quantity(100), Quantity(300)])

# ==============================================================================
# Test encode-decode of empty arrays


@insertTest(helper_name='write_read_attr',
            attrname='short_spectrum',
            setvalue=Quantity(numpy.empty(0, dtype='int16'), 'km'),
            expected=dict(rvalue=Quantity([], 'mm'),
                          type=DataType.Integer,
                          unit="mm"
                          ),
            expected_attrv=dict(rvalue=Quantity([], 'mm'),
                                type=DataType.Integer,
                                unit="mm"
                                ),
            expectedshape=(0,)
            )
@insertTest(helper_name='write_read_attr',
            attrname='short_image',
            setvalue=Quantity(numpy.empty((0, 0), dtype='int16'), 'mm'),
            expected=dict(rvalue=Quantity([], 'mm'),
                          type=DataType.Integer
                          ),
            expectedshape=(0, 0)
            )
@insertTest(helper_name='write_read_attr',
            attrname='boolean_spectrum',
            setvalue=numpy.empty(0, dtype='bool8'),
            expected=dict(type=DataType.Boolean),
            expectedshape=(0,)
            )
@insertTest(helper_name='write_read_attr',
            attrname='boolean_image',
            setvalue=numpy.empty((0, 0), dtype='bool8'),
            expected=dict(type=DataType.Boolean),
            expectedshape=(0, 0)
            )
# # Cannot test: Tango ignores writes of empty list for string attributes.
# @insertTest(helper_name='write_read_attr',
#             attrname='string_spectrum',
#             setvalue=[],
#             expected=dict(value=(), type=DataType.String,
#                           rvalue=(), wvalue=()))
#
# # Cannot test: Tango ignores writes of empty list for string attributes.
# @insertTest(helper_name='write_read_attr',
#             attrname='string_image',
#             setvalue=[[]],
#             expected=dict(value=[[]], type=DataType.String))
# ==============================================================================
# Test encode-decode of strings, booleans and uchars
@insertTest(helper_name='write_read_attr',
            attrname='uchar_image',
            setvalue=_UCHAR_IMG,
            expected=dict(rvalue=_UCHAR_IMG,
                          wvalue=_UCHAR_IMG,
                          type=DataType.Bytes,
                          label='uchar_image',
                          writable=True,
                          ),
            expected_attrv=dict(rvalue=_UCHAR_IMG,
                                value=_UCHAR_IMG,
                                wvalue=_UCHAR_IMG,
                                w_value=_UCHAR_IMG,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='uchar_spectrum',
            setvalue=_UCHAR_SPE,
            expected=dict(rvalue=_UCHAR_SPE,
                          wvalue=_UCHAR_SPE,
                          type=DataType.Bytes,
                          writable=True,
                          ),
            expected_attrv=dict(rvalue=_UCHAR_SPE,
                                value=_UCHAR_SPE,
                                wvalue=_UCHAR_SPE,
                                w_value=_UCHAR_SPE,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='uchar_scalar',
            setvalue='a',
            expected=dict(rvalue='a',
                          wvalue='a',
                          type=DataType.Bytes,
                          writable=True,
                          range=[None, None],
                          alarms=[None, None],
                          warnings=[None, None]
                          ),
            expected_attrv=dict(rvalue='a',
                                value='a',
                                wvalue='a',
                                w_value='a',
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='uchar_image',
            setvalue=_UINT8_IMG,
            expected=dict(rvalue=_UCHAR_IMG,
                          wvalue=_UCHAR_IMG,
                          type=DataType.Bytes,
                          label='uchar_image',
                          writable=True,
                          ),
            expected_attrv=dict(rvalue=_UCHAR_IMG,
                                value=_UCHAR_IMG,
                                wvalue=_UCHAR_IMG,
                                w_value=_UCHAR_IMG,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='uchar_spectrum',
            setvalue=_UINT8_SPE,
            expected=dict(rvalue=_UCHAR_SPE,
                          wvalue=_UCHAR_SPE,
                          type=DataType.Bytes,
                          writable=True,
                          ),
            expected_attrv=dict(rvalue=_UCHAR_SPE,
                                value=_UCHAR_SPE,
                                wvalue=_UCHAR_SPE,
                                w_value=_UCHAR_SPE,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='uchar_scalar',
            setvalue=97,  # ord('a')-->97
            expected=dict(rvalue='a',
                          wvalue='a',
                          type=DataType.Bytes,
                          writable=True,
                          range=[None, None],
                          alarms=[None, None],
                          warnings=[None, None]
                          ),
            expected_attrv=dict(rvalue='a',
                                value='a',
                                wvalue='a',
                                w_value='a',
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='string_image',
            setvalue=((_STR,) * 3,) * 2,
            expected=dict(rvalue=((_STR,) * 3,) * 2,
                          wvalue=((_STR,) * 3,) * 2,
                          type=DataType.String,
                          label='string_image',
                          writable=True,
                          ),
            expected_attrv=dict(value=((_STR,) * 3,) * 2,
                                w_value=((_STR,) * 3,) * 2,
                                rvalue=((_STR,) * 3,) * 2,
                                wvalue=((_STR,) * 3,) * 2,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='string_spectrum',
            setvalue=(_STR,) * 3,
            expected=dict(rvalue=(_STR,) * 3,
                          wvalue=(_STR,) * 3,
                          type=DataType.String,
                          label='string_spectrum',
                          writable=True,
                          ),
            expected_attrv=dict(value=(_STR,) * 3,
                                w_value=(_STR,) * 3,
                                rvalue=(_STR,) * 3,
                                wvalue=(_STR,) * 3,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='string_scalar',
            setvalue=_STR,
            expected=dict(rvalue=_STR,
                          wvalue=_STR,
                          type=DataType.String,
                          label='string_scalar',
                          writable=True,
                          ),
            expected_attrv=dict(value=_STR,
                                w_value=_STR,
                                rvalue=_STR,
                                wvalue=_STR,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='boolean_image',
            setvalue=_BOOL_IMG,
            expected=dict(rvalue=_BOOL_IMG,
                          wvalue=_BOOL_IMG,
                          type=DataType.Boolean,
                          label='boolean_image',
                          writable=True,
                          ),
            expected_attrv=dict(rvalue=_BOOL_IMG,
                                value=_BOOL_IMG,
                                wvalue=_BOOL_IMG,
                                w_value=_BOOL_IMG,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='boolean_spectrum',
            setvalue=_BOOL_SPE,
            expected=dict(rvalue=_BOOL_SPE,
                          wvalue=_BOOL_SPE,
                          type=DataType.Boolean,
                          writable=True,
                          ),
            expected_attrv=dict(rvalue=_BOOL_SPE,
                                value=_BOOL_SPE,
                                wvalue=_BOOL_SPE,
                                w_value=_BOOL_SPE,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='boolean_scalar',
            setvalue=False,
            expected=dict(rvalue=False,
                          wvalue=False,
                          type=DataType.Boolean,
                          writable=True,
                          range=[None, None],
                          alarms=[None, None],
                          warnings=[None, None]
                          ),
            expected_attrv=dict(rvalue=False,
                                value=False,
                                wvalue=False,
                                w_value=False,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
# ==============================================================================
# Test encode-decode with quantitiy conversions
@insertTest(helper_name='write_read_attr',
            attrname='float_image',
            setvalue=Quantity(_FLOAT_IMG, 'm'),
            expected=dict(rvalue=Quantity(_FLOAT_IMG, 'm'),
                          wvalue=Quantity(_FLOAT_IMG, 'm'),
                          type=DataType.Float,
                          writable=True,
                          quality=AttrQuality.ATTR_VALID,
                          label='float_image',
                          range=[Quantity(float('-inf'), 'mm'),
                                 Quantity(float('inf'), 'mm')],
                          alarms=[Quantity(float('-inf'), 'mm'),
                                  Quantity(float('inf'), 'mm')],
                          warnings=[Quantity(float('-inf'), 'mm'),
                                    Quantity(float('inf'), 'mm')]
                          ),
            expected_attrv=dict(rvalue=Quantity(_FLOAT_IMG, 'm'),
                                value=1000 * _FLOAT_IMG,
                                wvalue=Quantity(_FLOAT_IMG, 'm'),
                                w_value=1000 * _FLOAT_IMG,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=numpy.shape(_FLOAT_IMG)
            )
@insertTest(helper_name='write_read_attr',
            attrname='float_spectrum',
            setvalue=Quantity(_FLOAT_SPE, 'm'),
            expected=dict(rvalue=Quantity(_FLOAT_SPE, 'm'),
                          wvalue=Quantity(_FLOAT_SPE, 'm'),
                          type=DataType.Float,
                          quality=AttrQuality.ATTR_VALID
                          ),
            expected_attrv=dict(rvalue=Quantity(_FLOAT_SPE, 'm'),
                                value=1000 * _FLOAT_SPE,
                                wvalue=Quantity(_FLOAT_SPE, 'm'),
                                w_value=1000 * _FLOAT_SPE,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=numpy.shape(_FLOAT_SPE)
            )
@insertTest(helper_name='write_read_attr',
            attrname='float_scalar',
            setvalue=Quantity(0.01, 'm'),
            expected=dict(rvalue=Quantity(0.01, 'm'),
                          wvalue=Quantity(10, 'mm'),
                          type=DataType.Float,
                          name='float_scalar',
                          range=[Quantity(-12.30, 'mm'),
                                 Quantity(12.30, 'mm')],
                          alarms=[Quantity(-6.15, 'mm'),
                                  Quantity(6.15, 'mm')],
                          warnings=[Quantity(-3.69, 'mm'),
                                    Quantity(3.69, 'mm')]
                          ),
            expected_attrv=dict(value=10,
                                rvalue=Quantity(0.01, 'm'),
                                w_value=10,
                                wvalue=Quantity(10, 'mm'),
                                quality=AttrQuality.ATTR_ALARM
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='float_scalar',
            setvalue=Quantity(0.004, 'm'),
            expected=dict(rvalue=Quantity(4, 'mm'),
                          wvalue=Quantity(4, 'mm')
                          ),
            expected_attrv=dict(rvalue=Quantity(4, 'mm'),
                                wvalue=Quantity(0.004, 'm'),
                                quality=AttrQuality.ATTR_WARNING
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='float_scalar',
            setvalue=Quantity(3, 'mm'),
            expected=dict(rvalue=Quantity(3, 'mm'),
                          wvalue=Quantity(3, 'mm'),
                          type=DataType.Float,
                          ),
            expected_attrv=dict(value=3,
                                w_value=3,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='double_scalar',
            setvalue=Quantity(0.01, 'm'),
            expected=dict(rvalue=Quantity(0.01, 'm'),
                          wvalue=Quantity(10, 'mm'),
                          type=DataType.Float,
                          name="double_scalar",
                          range=[Quantity(-12.30, 'mm'),
                                 Quantity(12.30, 'mm')],
                          alarms=[Quantity(-6.15, 'mm'),
                                  Quantity(6.15, 'mm')],
                          warnings=[Quantity(-3.69, 'mm'),
                                    Quantity(3.69, 'mm')]
                          ),
            expected_attrv=dict(value=10,
                                rvalue=Quantity(0.01, 'm'),
                                w_value=10,
                                wvalue=Quantity(10, 'mm'),
                                quality=AttrQuality.ATTR_ALARM
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='double_scalar',
            setvalue=Quantity(0.004, 'm'),
            expected=dict(rvalue=Quantity(4, 'mm'),
                          wvalue=Quantity(4, 'mm')
                          ),
            expected_attrv=dict(rvalue=Quantity(4, 'mm'),
                                wvalue=Quantity(0.004, 'm'),
                                quality=AttrQuality.ATTR_WARNING
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='double_scalar',
            setvalue=Quantity(3, 'mm'),
            expected=dict(rvalue=Quantity(3, 'mm'),
                          wvalue=Quantity(3, 'mm'),
                          type=DataType.Float,
                          ),
            expected_attrv=dict(value=3,
                                w_value=3,
                                quality=AttrQuality.ATTR_VALID
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='short_image',
            setvalue=Quantity(_INT_IMG, 'm'),
            expected=dict(rvalue=Quantity(_INT_IMG, 'm'),
                          wvalue=Quantity(_INT_IMG, 'm'),
                          type=DataType.Integer
                          ),
            expected_attrv=dict(value=1000 * _INT_IMG,
                                rvalue=Quantity(_INT_IMG, 'm'),
                                wvalue=Quantity(_INT_IMG, 'm'),
                                w_value=1000 * _INT_IMG,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=numpy.shape(_INT_IMG)
            )
@insertTest(helper_name='write_read_attr',
            attrname='short_spectrum',
            setvalue=Quantity(_INT_SPE, 'm'),
            expected=dict(rvalue=Quantity(_INT_SPE, 'm'),
                          wvalue=Quantity(_INT_SPE, 'm'),
                          type=DataType.Integer),
            expectedshape=numpy.shape(_INT_SPE),
            expected_attrv=dict(value=1000 * _INT_SPE,
                                w_value=1000 * _INT_SPE
                                )
            )
@insertTest(helper_name='write_read_attr',
            attrname='short_scalar',
            setvalue=Quantity(1, 'm'),
            expected=dict(rvalue=Quantity(1000, 'mm'),
                          wvalue=Quantity(1000, 'mm'),
                          type=DataType.Integer,
                          label='short_scalar',
                          data_format=DataFormat._0D,
                          writable=True,
                          range=[Quantity(-1230, 'mm'),
                                 Quantity(1230, 'mm')],
                          alarms=[Quantity(-615, 'mm'),
                                  Quantity(615, 'mm')],
                          warnings=[Quantity(-369, 'mm'),
                                    Quantity(369, 'mm')]
                          ),
            expected_attrv=dict(value=1000,
                                w_value=1000,
                                quality=AttrQuality.ATTR_ALARM
                                )
            )
# ==============================================================================
# Test read of tango attributes
@insertTest(helper_name='write_read_attr',
            attrname='string_image_ro',
            expected=dict(rvalue=(('hello world',) * 3,) * 3,
                          wvalue=None,
                          type=DataType.String
                          ),
            expected_attrv=dict(value=(('hello world',) * 3,) * 3,
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=(3, 3),
            test_skip='Known (Py)Tango bug for empty string arrays',
            )
@insertTest(helper_name='write_read_attr',
            attrname='string_spectrum_ro',
            expected=dict(rvalue=('hello world',) * 3,
                          wvalue=None,
                          type=DataType.String),
            expected_attrv=dict(value=('hello world',) * 3,
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=(3,),
            test_skip='Known (Py)Tango bug for empty string arrays',
            )
@insertTest(helper_name='write_read_attr',
            attrname='string_scalar_ro',
            expected=dict(rvalue='hello world',
                          wvalue=None,
                          type=DataType.String
                          ),
            expected_attrv=dict(value='hello world',
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            )
@insertTest(helper_name='write_read_attr',
            attrname='boolean_image_ro',
            expected=dict(rvalue=numpy.ones((3, 3), dtype='b'),
                          wvalue=None,
                          type=DataType.Boolean
                          ),
            expected_attrv=dict(value=numpy.ones((3, 3), dtype='b'),
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=(3, 3))
@insertTest(helper_name='write_read_attr',
            attrname='boolean_spectrum_ro',
            expected=dict(rvalue=numpy.array([True] * 3),
                          wvalue=None,
                          type=DataType.Boolean,
                          label='boolean_spectrum_ro'
                          ),
            expected_attrv=dict(value=numpy.array([True] * 3),
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=(3,))
@insertTest(helper_name='write_read_attr',
            attrname='boolean_scalar_ro',
            expected=dict(rvalue=True,
                          wvalue=None,
                          type=DataType.Boolean,
                          range=[None, None],
                          alarms=[None, None],
                          warnings=[None, None],
                          data_format=DataFormat._0D,
                          writable=False
                          ),
            expected_attrv=dict(value=True,
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            )
@insertTest(helper_name='write_read_attr',
            attrname='float_image_ro',
            expected=dict(rvalue=Quantity([[1.23] * 3] * 3, 'mm'),
                          wvalue=None,
                          type=DataType.Float
                          ),
            expected_attrv=dict(value=[[1.23] * 3] * 3,
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID,
                                wvalue=None),
            expectedshape=(3, 3))
@insertTest(helper_name='write_read_attr',
            attrname='float_spectrum_ro',
            expected=dict(rvalue=Quantity([1.23] * 3, 'mm'),
                          wvalue=None,
                          type=DataType.Float
                          ),
            expected_attrv=dict(value=[1.23] * 3,
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            expectedshape=(3,)
            )
@insertTest(helper_name='write_read_attr',
            attrname='float_scalar_ro',
            expected=dict(rvalue=Quantity(1.23, 'mm'),
                          wvalue=None,
                          type=DataType.Float,
                          writable=False,
                          range=[Quantity(float('-inf'), 'mm'),
                                 Quantity(float('inf'), 'mm')],
                          alarms=[Quantity(float('-inf'), 'mm'),
                                  Quantity(float('inf'), 'mm')],
                          warnings=[Quantity(float('-inf'), 'mm'),
                                    Quantity(float('inf'), 'mm')]
                          ),
            expected_attrv=dict(value=1.23,
                                w_value=None,
                                quality=AttrQuality.ATTR_VALID
                                ),
            )
@insertTest(helper_name='write_read_attr',
            attrname='short_image_ro',
            expected=dict(rvalue=Quantity([[123] * 3] * 3, 'mm'),
                          wvalue=None,
                          type=DataType.Integer
                          ),
            expected_attrv=dict(rvalue=Quantity([[123] * 3] * 3, 'mm'),
                                value=[[123] * 3] * 3,
                                quality=AttrQuality.ATTR_VALID,
                                wvalue=None,
                                w_value=None),
            expectedshape=(3, 3)
            )
@insertTest(helper_name='write_read_attr',
            attrname='short_spectrum_ro',
            expected=dict(rvalue=Quantity([123] * 3, 'mm'),
                          wvalue=None,
                          type=DataType.Integer,
                          data_format=DataFormat._1D,
                          writable=False
                          ),
            expected_attrv=dict(rvalue=Quantity([123] * 3, 'mm'),
                                value=[123] * 3,
                                quality=AttrQuality.ATTR_VALID,
                                wvalue=None,
                                w_value=None),
            expectedshape=(3,)
            )
@insertTest(helper_name='write_read_attr',
            attrname='short_scalar_ro',
            expected=dict(rvalue=Quantity(123, 'mm'),
                          wvalue=None,
                          type=DataType.Integer,
                          data_format=DataFormat._0D,
                          writable=False,
                          range=[Quantity(float('-inf'), 'mm'),
                                 Quantity(float('inf'), 'mm')],
                          alarms=[Quantity(float('-inf'), 'mm'),
                                  Quantity(float('inf'), 'mm')],
                          warnings=[Quantity(float('-inf'), 'mm'),
                                    Quantity(float('inf'), 'mm')]
                          ),
            expected_attrv=dict(rvalue=Quantity(123, 'mm'),
                                value=123,
                                quality=AttrQuality.ATTR_VALID,
                                wvalue=None,
                                w_value=None)
            )
class AttributeTestCase(TangoSchemeTestLauncher, unittest.TestCase):
    """TestCase for the taurus.Attribute helper"""

    def tearDown(self):
        TangoSchemeTestLauncher.tearDown(self)

    def _getDecodePyTangoAttr(self, attr_name, cfg):
        """Helper for decode the PyTango attribute infoex
        """
        dev = PyTango.DeviceProxy(self.DEV_NAME)
        infoex = dev.get_attribute_config_ex(attr_name)[0]
        unit = unit_from_tango(infoex.unit)
        if cfg in ['range', 'alarms', 'warnings']:
            if cfg == 'range':
                low = infoex.min_value
                high = infoex.max_value
            elif cfg == 'alarms':
                low = infoex.alarms.min_alarm
                high = infoex.alarms.max_alarm
            elif cfg == 'warnings':
                low = infoex.alarms.min_warning
                high = infoex.alarms.max_warning
            if low == 'Not specified':
                low = '-inf'
            if high == 'Not specified':
                high = 'inf'
            return [Quantity(float(low), unit),
                    Quantity(float(high), unit)]
        elif cfg == 'label':
            return infoex.label
        else:
            return None

    def write_read_conf(self, attr_name, cfg, value, expected):
        """ Helper for checking the write-and-read of the Tango
        attribute configuration (existing in Taurus).
        """
        attr_fullname = '%s/%s' % (self.DEV_NAME, attr_name)
        attr = taurus.Attribute(attr_fullname)
        # write the property
        setattr(attr, cfg, value)
        # read the property
        got = getattr(attr, cfg)
        msg = '%s.%s from Taurus do not mach, expected %s read %s' %\
              (attr_name, cfg, expected, got)
        map(self.__assertValidValue, got, expected, msg)

        msg = '%s.%s from Tango do not mach, expected %s read %s' %\
              (attr_name, cfg, expected, got)
        tangovalue = self._getDecodePyTangoAttr(attr_name, cfg)
        map(self.__assertValidValue, got, tangovalue, msg)

    def write_read_attr(self, attrname=None, setvalue=None, expected=None,
                        expected_attrv=None, expectedshape=None):
        """check creation and correct write-and-read of an attribute"""

        if expected is None:
            expected = {}
        if expected_attrv is None:
            expected_attrv = {}

        name = '%s/%s' % (self.DEV_NAME, attrname)
        a = taurus.Attribute(name)

        if setvalue is None:
            read_value = a.read()
        else:
            read_value = a.write(setvalue,  with_read=True)

        msg = ('read() for "%s" did not return a TangoAttrValue (got a %s)' %
               (attrname, read_value.__class__.__name__))
        self.assertTrue(isinstance(read_value, TangoAttrValue), msg)

        # Test attribute
        for k, exp in expected.iteritems():
            try:
                got = getattr(a, k)
            except AttributeError:
                msg = ('The attribute, "%s" does not provide info on %s' %
                       (attrname, k))
                self.fail(msg)
            msg = ('%s for "%s" should be %r (got %r)' %
                   (k, attrname,  exp, got))
            self.__assertValidValue(exp, got, msg)

        # Test attribute value
        for k, exp in expected_attrv.iteritems():
            try:
                got = getattr(read_value, k)
            except AttributeError:
                msg = ('The read value for "%s" does not provide info on %s' %
                       (attrname, k))
                self.fail(msg)
            msg = ('%s for the value of %s should be %r (got %r)' %
                   (k, attrname,  exp, got))
            self.__assertValidValue(exp, got, msg)

        if expectedshape is not None:
            msg = ('rvalue.shape for %s should be %r (got %r)' %
                   (attrname, expectedshape, read_value.rvalue.shape))
            self.assertEqual(read_value.rvalue.shape, expectedshape, msg)

    def __assertValidValue(self, exp, got, msg):
        # if we are dealing with quantities, use the magnitude for comparing
        if isinstance(got, Quantity):
            got = got.to(Quantity(exp).units).magnitude
        if isinstance(exp, Quantity):
            exp = exp.magnitude
        try:
            # for those values that can be handled by numpy.allclose()
            chk = numpy.allclose(got, exp)
        except:
            if isinstance(got, numpy.ndarray):
                # uchars were not handled with allclose
                # UGLY!! but numpy.all does not work
                chk = got.tolist() == exp.tolist()
            else:
                # for the rest
                chk = bool(got == exp)

        self.assertTrue(chk, msg)

if __name__ == '__main__':
    pass



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

'''
Examples on using the evaluation scheme for exposing arbitrary non-tango quantities as taurus attributes
'''

__all__ = ['IcepapDriverParam']

from taurus.core.evaluation import EvaluationDevice
import re


class IcepapDriverParam(EvaluationDevice):
    '''A simple example of usage of the evaluation scheme for
    creating an icepap connection device to obtain icepap driver values.

    Important: note that only those members listed in `_symbols` will be available
    '''
    _symbols = ['getAxisParam']

    def __init__(self, *args, **kwargs):
        ''' Get from Database info the icepap host and port to connect. '''
        self.call__init__(EvaluationDevice, *args, **kwargs)

        # Get the icepap host and port to connect
        self.ipap = None
        import pyIcePAP

        try:
            db_name = self.getNameValidator().getDBName(self._full_name)
            db_name = db_name.replace('eval://', '')
            db_name = db_name.replace('db=', '')
            host, port = db_name.split(':')
            self.ipap = pyIcePAP.EthIcePAP(host, port)
            self.ipap.connect()
        except:
            pass

    def getAxisParam(self, axis, param):
        ''' return the axis parameter value. '''
        if self.ipap is None or not self.ipap.connected:
            raise Exception('Not a valid icepap connection')

        try:
            value = self.ipap.readParameter(axis, param)
            return float(value)
        except:
            return value


#=========================================================================
# Just for testing
#=========================================================================

ATTR_IPAP_POS = 'eval://db=icepap06:5000;dev=taurus.core.evaluation.ipap_example.IcepapDriverParam;getAxisParam(1,"POS")'


def _test1():
    import taurus.core
    a = taurus.Attribute(ATTR_IPAP_POS)
    print "axi_s pos:", a.read().value


def _test2():
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.qt.qtgui.display import TaurusLabel
    app = TaurusApplication()

    tl = TaurusLabel()
    tl.setModel(ATTR_IPAP_POS)
    tl.show()

    sys.exit(app.exec_())

if __name__ == "__main__":
    _test1()
    _test2()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module parses jdraw files"""

from __future__ import absolute_import

__all__ = ["new_parser", "parse"]

import os
import re
import imp

from ply import lex
from ply import yacc

from taurus.core.util.log import Logger

tokens = ('NUMBER', 'SYMBOL', 'LBRACKET', 'RBRACKET', 'TWOP', 'COMMA',
          'JDFILE', 'GLOBAL', 'JDLINE', 'JDRECTANGLE', 'JDROUNDRECTANGLE',
          'JDGROUP', 'JDELLIPSE', 'JDBAR', 'JDSWINGOBJECT', 'JDLABEL', 'JDPOLYLINE',
          'JDIMAGE', 'JDAXIS', 'JDSLIDER', 'JDSPLINE', 'TEXT',
          'true', 'false',
          )

t_LBRACKET = r'\{'
t_RBRACKET = r'\}'
t_TWOP = r'\:'
t_COMMA = r'\,'

t_TEXT = r'\"[^"]*\"'

reserved = {
    'JDFile': 'JDFILE',
    'Global': 'GLOBAL',
    'JDLine': 'JDLINE',
    'JDRectangle': 'JDRECTANGLE',
    'JDRoundRectangle': 'JDROUNDRECTANGLE',
    'JDGroup': 'JDGROUP',
    'JDEllipse': 'JDELLIPSE',
    'JDBar': 'JDBAR',
    'JDSwingObject': 'JDSWINGOBJECT',
    'JDLabel': 'JDLABEL',
    'JDPolyline': 'JDPOLYLINE',
    'JDImage': 'JDIMAGE',
    'JDAxis': 'JDAXIS',
    'JDSlider': 'JDSLIDER',
    'JDSpline': 'JDSPLINE',
    'true': 'true',
    'false': 'false',

}


def t_SYMBOL(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value, 'SYMBOL')    # Check for reserved words
    return t


def t_NUMBER(t):
    r'[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?'
    try:
        t.value = float(t.value)
    except:
        t.lexer.log.info("[%d]: Number %s is not valid!" % (t.lineno, t.value))
        t.value = 0
    return t

# Ignored characters
t_ignore = " \t"


def t_newline(t):
    r'\n+'
    t.lexer.lineno += t.value.count("\n")


def t_error(t):
    t.lexer.log.info("[%d]: Illegal character '%s'" %
                     (t.lexer.lineno, t.value[0]))
    t.lexer.skip(1)


def p_error(p):
    p.lexer.log.error("[%d]: Syntax error in input [%s]" %
                      (p.lexer.lineno, (str(p))))

#-------------------------------------------------------------------------
# Yacc Starting symbol
#-------------------------------------------------------------------------


def p_jdfile(p):
    ''' jdfile :  JDFILE SYMBOL LBRACKET global element_list RBRACKET '''
    factory = p.parser.factory
    p[0] = factory.getSceneObj(p[5])

    if p[0] is None:
        p.parser.log.info("[%d]: Unable to create Scene" % p.lexer.lineno)


def p_jdfile_empty(p):
    ''' jdfile : JDFILE LBRACKET global RBRACKET '''
    factory = p.parser.factory
    p[0] = factory.getSceneObj([])

    if p[0] is None:
        p.parser.log.info("[%d]: Unable to create Scene" % p.lexer.lineno)


def p_global(p):
    ''' global : GLOBAL LBRACKET RBRACKET
               | GLOBAL LBRACKET parameter_list RBRACKET'''
    if len(p) == 4:
        p[0] = {}
    else:
        p[0] = p[3]


def p_element_list(p):
    '''element_list : element_list element '''
    if not p[2] is None:
        p[0] = p[1] + [p[2]]
    else:
        p[0] = p[1]


def p_element(p):
    '''element_list : element '''
    p[0] = [p[1]]


def p_single_element(p):
    '''element : obj LBRACKET RBRACKET
               | obj LBRACKET parameter_list RBRACKET'''

    #elems = None
    if len(p) == 4:
        p[3] = {}

    # determine the model name
    org = name = p[3].get('name')

    # change name: if item has no name it should take the name of its nearest-named-parent
    # as no-name it is also considered any element with a name like a reserved
    # keyword or plain value
    keywords = ['JDGroup'] + [n.replace('JD', '') for n in reserved]
    if not name or name in keywords or re.match('[0-9]+$', name):
        #if re.match('[0-9]+$',name):  p[3]['mapvalue'] = name
        p[3]['name'] = name = ""
        # print '\t name is empty, checking stack: %s' % [str(s) for s in
        # reversed(p.parser.modelStack)]
        for model in reversed(p.parser.modelStack):
            if model and model not in keywords and not re.match('[0-9]+$', model):
                # print 'Setting Model %s to %s' % (model,p[1])
                p[3]['name'] = name = model
                break
    # print 'parser: %s => %s [%s]' % (org,name,p.parser.modelStack)
    extension = p[3].get("extensions")
    if p.parser.modelStack2:
        if extension is None:
            p[3]["extensions"] = p.parser.modelStack2[0]
        elif len(p.parser.modelStack2) == 2:
            extension.update(p.parser.modelStack2[0])
            p[3]["extensions"] = extension

    # create the corresponding element
    factory = p.parser.factory
    #p.parser.log.debug('ret = factory.getObj(%s,%s)'% (str(p[1]),str(p[3])))
    ret = factory.getObj(p[1], p[3])

    if ret is None:
        p.parser.log.info("[%d]: Unable to create obj '%s'" %
                          (p.lexer.lineno, p[1]))

    # clear the model stack
    # if name:
    p.parser.modelStack.pop()
    if extension:
        p.parser.modelStack2.pop()

    #extension = p[3].get("extensions")
    p[0] = ret


def p_obj(p):
    '''obj : JDLINE
           | JDRECTANGLE
           | JDROUNDRECTANGLE
           | JDGROUP
           | JDELLIPSE
           | JDBAR
           | JDSWINGOBJECT
           | JDLABEL
           | JDPOLYLINE
           | JDIMAGE
           | JDAXIS
           | JDSLIDER
           | JDSPLINE'''
    p[0] = p[1]


def p_parameter_list(p):
    '''parameter_list : parameter_list parameter'''
    p[0] = p[1]
    p[0].update(p[2])


def p_parameter(p):
    '''parameter_list : parameter'''
    p[0] = p[1]


def p_single_parameter(p):
    '''parameter : SYMBOL TWOP param_value'''
    # modelStacks added in this method control how extensions/models/names are
    # propagated within JDGroup items
    if p[1] == 'name':
        p.parser.modelStack.append(p[3])
    if p[1] == 'extensions':
        p.parser.modelStack2.append(p[3])
    p[0] = {p[1]: p[3]}


def p_complex_parameter(p):
    '''parameter : SYMBOL TWOP LBRACKET RBRACKET
                 | SYMBOL TWOP LBRACKET parameter_list RBRACKET
                 | SYMBOL TWOP LBRACKET element_list RBRACKET'''
    if len(p) == 5:
        p[4] = []
    p[0] = {p[1]: p[4]}
    if p[1] == 'extensions':
        p.parser.modelStack2.append(p[4])


def p_param_value_number_list(p):
    '''param_value : value_list '''
    if len(p[1]) == 1:
        p[0] = p[1][0]
    else:
        p[0] = p[1]


def p_value_list(p):
    ''' value_list : value_list COMMA value '''
    p[0] = p[1] + [p[3]]


def p_value_list_value(p):
    ''' value_list : value '''
    p[0] = [p[1]]


def p_value_number(p):
    '''value : NUMBER'''
    p[0] = p[1]


def p_value_text(p):
    '''value : TEXT'''
    p[0] = p[1].strip('"')


def p_value_bool(p):
    '''value : true
             | false'''
    p[0] = p[1] == 'true'


def new_parser(optimize=None, debug=0, outputdir=None):
    log = Logger('JDraw Parser')

    if optimize is None:
        from taurus import tauruscustomsettings
        optimize = getattr(tauruscustomsettings, 'PLY_OPTIMIZE', 1)
    if outputdir is None:
        # use '.taurus' dir in the user "home" dir
        outputdir = os.path.join(os.path.expanduser('~'), '.taurus')
        # create the output dir if it didn't exist
        if not os.path.exists(outputdir):
            os.makedirs(outputdir)

    debuglog = None
    if debug:
        debuglog = log

    common_kwargs = dict(optimize=optimize, outputdir=outputdir,
                         debug=debug, debuglog=debuglog, errorlog=log)

    # lex/yacc v<3.0 do not accept  debuglog or errorlog keyword args
    if int(lex.__version__.split('.')[0]) < 3:
        common_kwargs.pop('debuglog')
        common_kwargs.pop('errorlog')

    try:
        _path = os.path.join(outputdir, 'jdraw_lextab.py')
        jdraw_lextab = imp.load_source('jdraw_lextab', _path)
    except:
        jdraw_lextab = 'jdraw_lextab'

    try:
        _path = os.path.join(outputdir, 'jdraw_yacctab.py')
        jdraw_yacctab = imp.load_source('jdraw_lextab', _path)
    except:
        jdraw_yacctab = 'jdraw_yacctab'

    # Lexer
    l = lex.lex(lextab=jdraw_lextab, **common_kwargs)

    # Yacc
    try:
        p = yacc.yacc(tabmodule=jdraw_yacctab, debugfile=None, write_tables=1,
                      **common_kwargs)
    except Exception, e:
        msg = ('Error creating jdraw parser.\n' +
               'HINT: Try removing jdraw_lextab.* and jdraw_yacctab.* from %s' %
               outputdir)
        raise RuntimeError(msg)

    return l, p


def parse(filename=None, factory=None):

    if filename is None or factory is None:
        return

    _, p = new_parser()

    p.factory = factory
    p.modelStack = []
    p.modelStack2 = []

    res = None
    try:
        filename = os.path.realpath(filename)
        f = open(filename)
        res = yacc.parse(f.read())
    except:
        log = Logger('JDraw Parser')
        log.warning("Failed to parse %s" % filename)
        log.debug("Details:", exc_info=1)
    return res


if __name__ == "__main__":
    new_parser()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides Qt color management for taurus"""

__all__ = ["Grabber", "grabWidget"]

__docformat__ = 'restructuredtext'

import time
import threading
import os.path

from taurus.external.qt import Qt
from taurus.core.util.log import Logger

_LOGGER = None


def _getLogger():
    global _LOGGER
    if _LOGGER is None:
        _LOGGER = Logger('Grabber')
    return _LOGGER


class GrabberThread(threading.Thread):

    def __init__(self, widget, fileName, period):
        threading.Thread.__init__(self, name="Grabber")
        self.daemon = True
        if period <= 0:
            raise ValueError("period MUST be greater than 0")
        self._period = period
        self._continue = True
        self._grabber = Grabber(widget, fileName)

    def run(self):
        period = self._period
        while self._continue:
            self._grabber.grabTrigger()
            time.sleep(period)

    def stop(self):
        self._continue = False


class Grabber(Qt.QObject, Logger):

    grab = Qt.pyqtSignal()

    def __init__(self, widget, fileName):
        Qt.QObject.__init__(self)
        Logger.__init__(self)
        self._widget = widget
        self._fileName = fileName
        self.grab.connect(self.onGrab)

    def grabTrigger(self):
        self.grab.emit()

    def onGrab(self):
        Grabber._grabWidget(self._widget, self._fileName)

    @staticmethod
    def _grabWidget(widget, fileName):
        _getLogger().debug("Grabbing widget to '%s':", fileName)
        try:
            pixmap = Qt.QPixmap.grabWidget(widget)
            if fileName.endswith('.svg'):
                from taurus.external.qt import QtSvg
                generator = QtSvg.QSvgGenerator()
                generator.setFileName(fileName)
                generator.setSize(pixmap.size())
                if hasattr(generator, 'setViewBox'):
                    viewBox = Qt.QRect(Qt.QPoint(0, 0), pixmap.size())
                    generator.setViewBox(viewBox)
                generator.setTitle("Taurus widget")
                generator.setDescription("An SVG drawing created by the taurus "
                                         "widget grabber")
                painter = Qt.QPainter()
                painter.begin(generator)
                try:
                    painter.drawPixmap(0, 0, -1, -1, pixmap)
                finally:
                    painter.end()
            else:
                pixmap.save(fileName, quality=100)
        except:
            _getLogger().warning("Could not save file into '%s':", fileName,
                                 exc_info=1)

    @staticmethod
    def grabWidget(widget, fileName, period=None):
        """Grabs the given widget into the given image filename. If period is
        not given (or given with None) means grab immediately once and return.
        If period is given and >0 means grab the image every period seconds

        .. warning:
            this method **MUST** be called from the same thread which created
            the widget

        :param widget: (Qt.QWidget) the qt widget to be grabbed
        :param fileName: (str) the name of the image file
        :param period: (float) period (seconds)
        """
        if period is None:
            return Grabber._grabWidget(widget, fileName)
        ret = GrabberThread(widget, fileName, period)
        ret.start()
        return ret


def grabWidget(widget, fileName, period=None):
    return Grabber.grabWidget(widget, fileName, period=period)

grabWidget.__doc__ = Grabber.grabWidget.__doc__



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides utilities for skipping certain sets of tests"""

#__all__ = []

__docformat__ = 'restructuredtext'

from taurus.external import unittest
from taurus import tauruscustomsettings
from taurus import Logger


def skipUnlessGui():
    '''Decorator to indicate that the given test should be skipped if GUI
    Tests are not enabled.

    It can be applied both to :class:`unittest.TestCase` classes and to
    test methods::

        class FooTest(unittest.TestCase):
            def test_something_which_does_not_need_gui()
                (...)

            @skipUnlessGui()
            def test_something_that requires_gui()
                (...)

        @skipUnlessGui()
        class GUITest(unittest.TestCase):
            (...)

    Note: using skipUnlessGui is equivalent to:

        @skipunless(taurus.test.GUI_TESTS_ENABLED, 'requires GUI')

    '''
    Logger.deprecated(dep='skipUnlessGui', rel='4.0',
                      alt='taurustestsuite --exclude-pattern')
    return unittest.skipUnless(GUI_TESTS_ENABLED, 'requires GUI')


def _hasgui():
    '''Returns True if GUI is available. False otherwise
    The current implementation is not very robust: it just looks for the
    'DISPLAY' environment variable on posix systems and assumes True
    for other systems'''
    import os
    if os.name == 'posix' and not os.getenv('DISPLAY'):
        return False
    else:
        return True

GUI_TESTS_ENABLED = getattr(
    tauruscustomsettings, 'ENABLE_GUI_TESTS', _hasgui())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This package contains a collection of taurus console widgets"""

__all__ = ["TaurusConsoleWidget"]

__docformat__ = 'restructuredtext'

from IPython.utils.traitlets import Unicode
try:
    from IPython.qt.console.rich_ipython_widget import RichIPythonWidget
except ImportError:
    from IPython.frontend.qt.console.rich_ipython_widget \
        import RichIPythonWidget

default_gui_banner = """\
Taurus console -- An enhanced IPython console for taurus.

?         -> Introduction and overview of IPython's features.
%quickref -> Quick reference.
help      -> Python's own help system.
object?   -> Details about 'object', use 'object??' for extra details.
%guiref   -> A brief reference about the graphical user interface.
"""


class TaurusConsoleWidget(RichIPythonWidget):

    banner = Unicode(config=True)

    #------ Trait default initializers ---------------------------------------
    def _banner_default(self):
        banner = default_gui_banner
        if 'TerminalInteractiveShell' in self.config:
            shell = self.config.TerminalInteractiveShell
            if 'banner' in shell:
                banner = shell.banner
            elif 'banner1' in shell and 'banner2' in shell:
                banner = "\n".join((shell.banner1, shell.banner2))
        return banner



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides base taurus tree item and a base tree model"""

__all__ = ["TaurusBaseTreeItem", "TaurusBaseModel", "TaurusBaseProxyModel"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.core.taurusbasetypes import TaurusElementType
from taurus.core.util.log import Logger

QtQt = Qt.Qt


class TaurusBaseTreeItem(object):
    """A generic node"""

    DisplayFunc = str

    def __init__(self, model, data, parent=None):
        self._model = model
        self._itemData = data
        self._parentItem = parent
        self._childItems = []
        self._depth = self._calcDepth()

    def itemData(self):
        """The internal itemData object

        :return: (object) object holding the data of this item
        """
        return self._itemData

    def depth(self):
        """Depth of the node in the hierarchy

        :return: (int) the node depth
        """
        return self._depth

    def appendChild(self, child):
        """Adds a new child node

        :param child: (TaurusTreeBaseItem) child to be added
        """
        self._childItems.append(child)

    def child(self, row):
        """Returns the child in the given row

        :return: (TaurusTreeBaseItem) the child node for the given row"""
        return self._childItems[row]

    def childCount(self):
        """Returns the number of childs for this node

        :return: (int) number of childs for this node
        """
        return len(self._childItems)

    def hasChildren(self):
        return len(self._childItems) > 0

    def data(self, index):
        """Returns the data of this node for the given index

        :return: (object) the data for the given index
        """
        return self._itemData[index.column()]

    def icon(self, index):
        return None

    def toolTip(self, index):
        return self.data(index)

    def setData(self, index, data):
        """Sets the node data

        :param data: (object) the data to be associated with this node
        """
        self._itemData = data

    def parent(self):
        """Returns the parent node or None if no parent exists

        :return: (TaurusTreeBaseItem) the parent node
        """
        return self._parentItem

    def row(self):
        """Returns the row for this node

        :return: (int) row number for this node
        """
        if self._parentItem is None:
            return 0
        return self._parentItem._childItems.index(self)

    def _calcDepth(self):
        d = 0
        n = self.parent()
        while n is not None:
            n = n.parent()
            d += 1
        return d

    def display(self):
        """Returns the display string for this node

        :return: (str) the node's display string"""
        if not hasattr(self, "_display"):
            if self._itemData is None:
                return None
            self._display = self.DisplayFunc(self._itemData)
        return self._display

    def qdisplay(self):
        """Returns the display QString for this node

        :return: (Qt.QString) the node's display string"""
        if not hasattr(self, "_qdisplay"):
            d = self.display()
            if d is None:
                return None
            self._qdisplay = Qt.QString(d)
        return self._qdisplay

    def mimeData(self, index):
        return self.data(index)

    def role(self):
        """Returns the prefered role for the item.
        This implementation returns taurus.core.taurusbasetypes.TaurusElementType.Unknown

        This method should be able to return any kind of python object as long
        as the model that is used is compatible.

        :return: (taurus.core.taurusbasetypes.TaurusElementType) the role in form of element type"""
        return TaurusElementType.Unknown

    def __str__(self):
        return self.display()


class TaurusBaseModel(Qt.QAbstractItemModel, Logger):
    """The base class for all Taurus Qt models."""

    ColumnNames = ()
    ColumnRoles = (),

    DftFont = Qt.QFont("Mono", 8)

    def __init__(self, parent=None, data=None):
        Qt.QAbstractItemModel.__init__(self, parent)
        Logger.__init__(self)
        # if qt < 4.6, beginResetModel and endResetModel don't exist. In this
        # case we set beginResetModel to be an empty function and endResetModel
        # to be reset.
        if not hasattr(Qt.QAbstractItemModel, "beginResetModel"):
            self.beginResetModel = lambda: None
            self.endResetModel = self.reset
        self._data_src = None
        self._rootItem = None
        self._filters = []
        self._selectables = [self.ColumnRoles[0][-1]]
        self.setDataSource(data)

    def __getattr__(self, name):
        return getattr(self.dataSource(), name)

    def createNewRootItem(self):
        return TaurusBaseTreeItem(self, self.ColumnNames)

    def refresh(self, refresh_source=False):
        self.beginResetModel()
        self._rootItem = self.createNewRootItem()
        self.setupModelData(self.dataSource())
        self.endResetModel()

    def setupModelData(self, data):
        raise NotImplementedError("setupModelData must be implemented "
                                  "in %s" % self.__class__.__name__)

    def roleIcon(self, role):
        raise NotImplementedError("roleIcon must be implemented "
                                  "in %s" % self.__class__.__name__)

    def roleSize(self, role):
        raise NotImplementedError("roleSize must be implemented "
                                  "in %s" % self.__class__.__name__)

    def roleToolTip(self, role):
        raise NotImplementedError("roleToolTip must be implemented "
                                  "in %s" % self.__class__.__name__)

    def setDataSource(self, data_src):
        self._data_src = data_src
        self.refresh()

    def dataSource(self):
        return self._data_src

    def setSelectables(self, seq_elem_types):
        self._selectables = seq_elem_types

    def selectables(self):
        return self._selectables

    def role(self, column, depth=0):
        cr = self.ColumnRoles
        if column == 0:
            return cr[0][depth]
        return self.ColumnRoles[column]

    def columnCount(self, parent=Qt.QModelIndex()):
        return len(self.ColumnRoles)

    def columnIcon(self, column):
        return self.roleIcon(self.role(column))

    def columnToolTip(self, column):
        return self.roleToolTip(self.role(column))

    def columnSize(self, column):
        role = self.role(column)
        s = self.roleSize(role)
        return s

    def pyData(self, index, role=QtQt.DisplayRole):
        if not index.isValid():
            return None

        item = index.internalPointer()

        ret = None
        if role == QtQt.DisplayRole or role == QtQt.EditRole:
            ret = item.data(index)
#        elif role == QtQt.CheckStateRole:
#            data = item.data(index)
#            if type(data) != bool:
#                data = str(data).lower() == 'true'
#            ret = QtQt.Unchecked
#            if data == True:
#                ret = QtQt.Checked
        elif role == QtQt.DecorationRole:
            ret = item.icon(index)
        elif role == QtQt.ToolTipRole:
            ret = item.toolTip(index)
        # elif role == QtQt.SizeHintRole:
        #    ret = self.columnSize(column)
        elif role == QtQt.FontRole:
            ret = self.DftFont
        elif role == QtQt.UserRole:
            ret = Qt.QVariant(item)
        return ret

    def data(self, index, role=QtQt.DisplayRole):
        ret = self.pyData(index, role)
        if ret is None:
            ret = Qt.QVariant()
        else:
            ret = Qt.QVariant(ret)
        return ret

    def _setData(self, index, qvalue, role=QtQt.EditRole):
        item = index.internalPointer()
        pyobj = Qt.from_qvariant(qvalue)
        if pyobj is NotImplemented:
            self.warning(
                "Failed attempt to convert a QValue. Maybe it is due to Qt<4.6")
        item.setData(index, pyobj)
        return True

    def flags(self, index):
        if not index.isValid():
            return 0

        ret = QtQt.ItemIsEnabled | QtQt.ItemIsDragEnabled

        item = index.internalPointer()
        column, depth = index.column(), item.depth()
        taurus_role = self.role(column, depth)

        if taurus_role in self.selectables():
            ret |= QtQt.ItemIsSelectable
        return ret

    def headerData(self, section, orientation, role=QtQt.DisplayRole):
        ret = None
        if orientation == QtQt.Horizontal:
            if role == QtQt.TextAlignmentRole:
                ret = int(QtQt.AlignLeft | QtQt.AlignVCenter)
            elif role == QtQt.DisplayRole:
                ret = self.ColumnNames[section]
            elif role == QtQt.SizeHintRole:
                ret = Qt.QSize(self.columnSize(section))
                ret.setHeight(24)
            elif role == QtQt.ToolTipRole:
                ret = self.columnToolTip(section)
            elif role == QtQt.DecorationRole:
                ret = self.columnIcon(section)

        return Qt.QVariant(ret)

    def index(self, row, column, parent=Qt.QModelIndex()):
        if not self.hasIndex(row, column, parent):
            return Qt.QModelIndex()
        if not parent.isValid():
            parentItem = self._rootItem
        else:
            parentItem = parent.internalPointer()
        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        return Qt.QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return Qt.QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem is None or parentItem == self._rootItem:
            return Qt.QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent=Qt.QModelIndex()):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parentItem = self._rootItem
        else:
            parentItem = parent.internalPointer()
        if parentItem is None:
            return 0
        return parentItem.childCount()

    def hasChildren(self, parent=Qt.QModelIndex()):
        if parent.column() > 0:
            return 0
        if not parent.isValid():
            parentItem = self._rootItem
        else:
            parentItem = parent.internalPointer()

        if parentItem is None:
            return False
        return parentItem.hasChildren()


class TaurusBaseProxyModel(Qt.QSortFilterProxyModel):
    """A taurus base Qt filter & sort model"""

    def __init__(self, parent=None):
        Qt.QSortFilterProxyModel.__init__(self, parent)

        # filter configuration
        self.setFilterCaseSensitivity(QtQt.CaseInsensitive)
        self.setFilterKeyColumn(0)
        self.setFilterRole(QtQt.DisplayRole)

        # sort configuration
        self.setSortCaseSensitivity(QtQt.CaseInsensitive)
        self.setSortRole(QtQt.DisplayRole)

        # general configuration
        self.setDynamicSortFilter(True)

    def __getattr__(self, name):
        return getattr(self.sourceModel(), name)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides base tree widget"""

__all__ = ["QBaseTreeWidget"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.qt.qtgui.model import QBaseModelWidget, BaseToolBar
from taurus.qt.qtgui.util import ActionFactory


class _NavigationWidget(Qt.QFrame):
    """Internal widget that provides a navigation path to be placed in a toolbar"""

    def __init__(self, treeWidget, toolBarWidget, parent=None):
        super(_NavigationWidget, self).__init__(parent)
        self._tree = treeWidget
        self._toolbar = toolBarWidget
        self._label = Qt.QLabel()
        l = Qt.QHBoxLayout()
        l.setContentsMargins(4, 0, 4, 0)
        self.setLayout(l)
        l.addWidget(self._label)

    def treeWidget(self):
        return self._tree

    def viewWidget(self):
        return self.treeWidget().viewWidget()

    def toolBarWidget(self):
        return self._toolbar

    def updateSelection(self, index):
        treeWidget = self.treeWidget()
        src_model = treeWidget.getBaseQModel()
        toolbar = self.toolBarWidget()
        txt = u""

        while index.isValid():
            src_index = treeWidget._mapToSource(index)
            name = src_model.pyData(src_index, Qt.Qt.DisplayRole)
            txt = u" \u00bb " + name + txt
            index = index.parent()
        txt = u"Root" + txt
        self._label.setText(txt)

    def onGotoNode(self, *args):
        label = self.sender()
        persistent_index = label.parent().index()
        index = Qt.QModelIndex(persistent_index)
        tree = self.viewWidget()
        tree.setRootIndex(index)
        tree.setCurrentIndex(index.child(0, 0))


class NavigationToolBar(BaseToolBar):

    goIntoTriggered = Qt.pyqtSignal()
    goUpTriggered = Qt.pyqtSignal()
    goTopTriggered = Qt.pyqtSignal()

    def __init__(self, view=None, parent=None, designMode=False):
        BaseToolBar.__init__(self, name="Taurus selection toolbar", view=view,
                             parent=parent, designMode=designMode)

        af = ActionFactory()
        self._goIntoAction = af.createAction(self, "Go Into",
                                             icon=Qt.QIcon.fromTheme("go-down"),
                                             tip="Go into the selected item",
                                             triggered=self.goInto)
        self._goUpAction = af.createAction(self, "Go Up",
                                           icon=Qt.QIcon.fromTheme("go-up"),
                                           tip="Go up one level",
                                           triggered=self.goUp)
        self._goTopAction = af.createAction(self, "Go Top",
                                            icon=Qt.QIcon.fromTheme("go-top"),
                                            tip="Go to top level",
                                            triggered=self.goTop)
        self.addAction(self._goIntoAction)
        self.addAction(self._goUpAction)
        self.addAction(self._goTopAction)
        self._navigationWidget = _NavigationWidget(view, self, parent=self)
        self._navigationAction = self.addWidget(self._navigationWidget)

    def goIntoAction(self):
        return self._goIntoAction

    def goTopAction(self):
        return self._goTopAction

    def goUpAction(self):
        return self._goUpAction

    def goInto(self):
        self.goIntoTriggered.emit()

    def goUp(self):
        self.goUpTriggered.emit()

    def goTop(self):
        self.goTopTriggered.emit()


class ExpansionBar(BaseToolBar):

    expandTriggered = Qt.pyqtSignal()
    collapseTriggered = Qt.pyqtSignal()
    expandSelectionTriggered = Qt.pyqtSignal()
    collapseSelectionTriggered = Qt.pyqtSignal()

    def __init__(self, view=None, parent=None, designMode=False):
        BaseToolBar.__init__(self, name="Taurus selection toolbar", view=view,
                             parent=parent, designMode=designMode)

        af = ActionFactory()
        self._expandAllAction = af.createAction(self, "Expand All",
                                                icon=Qt.QIcon(
                                                    "actions:expand.png"),
                                                tip="Expand all items",
                                                triggered=self.onExpandAll)
        self._collapseAllAction = af.createAction(self, "Collapse All",
                                                  icon=Qt.QIcon(
                                                      "actions:collapse.png"),
                                                  tip="Collapse all items",
                                                  triggered=self.onCollapseAll)
        self._expandSelectionAction = af.createAction(self, "Expand selection",
                                                      icon=Qt.QIcon(
                                                          "actions:expand-selection.png"),
                                                      tip="Expand selection",
                                                      triggered=self.onExpandSelection)
        self._collapseSelectionAction = af.createAction(self, "Collapse All",
                                                        icon=Qt.QIcon(
                                                            "actions:collapse-selection.png"),
                                                        tip="Collapse selection",
                                                        triggered=self.onCollapseSelection)
        self.addAction(self._expandAllAction)
        self.addAction(self._collapseAllAction)
        self.addAction(self._expandSelectionAction)
        self.addAction(self._collapseSelectionAction)

    def onExpandAll(self):
        self.expandTriggered.emit()

    def onCollapseAll(self):
        self.collapseTriggered.emit()

    def onExpandSelection(self):
        self.expandSelectionTriggered.emit()

    def onCollapseSelection(self):
        self.collapseSelectionTriggered.emit()


class QBaseTreeWidget(QBaseModelWidget):
    """A pure Qt tree widget implementing a tree with a navigation toolbar"""

    def __init__(self, parent=None, designMode=False, with_navigation_bar=True,
                 with_filter_widget=True, with_selection_widget=True,
                 with_refresh_widget=True, perspective=None, proxy=None):

        if with_navigation_bar:
            if isinstance(with_navigation_bar, (bool, int)):
                self._with_navigation_bar = NavigationToolBar
            else:
                self._with_navigation_bar = with_navigation_bar
        else:
            self._with_navigation_bar = None

        QBaseModelWidget.__init__(self, parent,
                                  with_filter_widget=with_filter_widget,
                                  with_selection_widget=with_selection_widget,
                                  with_refresh_widget=with_refresh_widget,
                                  perspective=perspective, proxy=proxy)

    def createToolArea(self):
        ta = QBaseModelWidget.createToolArea(self)

        e_bar = self._expandBar = ExpansionBar(view=self, parent=self)
        e_bar.expandTriggered.connect(self.expandAllTree)
        e_bar.collapseTriggered.connect(self.collapseAllTree)
        e_bar.expandSelectionTriggered.connect(self.expandSelectionTree)
        e_bar.collapseSelectionTriggered.connect(self.collapseSelectionTree)
        ta.append(e_bar)

        if self._with_navigation_bar:
            n_bar = self._navigationToolBar = self._with_navigation_bar(
                view=self, parent=self)
            n_bar.goIntoTriggered.connect(self.goIntoTree)
            n_bar.goTopTriggered.connect(self.goTopTree)
            n_bar.goUpTriggered.connect(self.goUpTree)
            ta.append(n_bar)
        else:
            self._navigationToolBar = None
        return ta

    def createViewWidget(self, klass=None):
        if klass is None:
            klass = Qt.QTreeView
        tree = klass()
        tree.setSortingEnabled(True)
        tree.sortByColumn(0, Qt.Qt.AscendingOrder)
        tree.setUniformRowHeights(True)
        tree.setAlternatingRowColors(True)
        tree.setSelectionBehavior(Qt.QAbstractItemView.SelectRows)
        tree.setSelectionMode(Qt.QAbstractItemView.ExtendedSelection)
        tree.setDragEnabled(True)
        tree.setDropIndicatorShown(True)

        tree.expanded.connect(self.onExpanded)
        tree.clicked.connect(self._onClicked)
        tree.doubleClicked.connect(self._onDoubleClicked)
        h = tree.header()
        h.setResizeMode(0, Qt.QHeaderView.Stretch)
        return tree

    def treeView(self):
        return self.viewWidget()

    def goIntoAction(self):
        return self._navigationToolBar.goIntoAction()

    def goTopAction(self):
        return self._navigationToolBar.goTopAction()

    def goUpAction(self):
        return self._navigationToolBar.goUpAction()

    def expandAllTree(self):
        self.statusBar().showMessage("Expanding all items... (it may take a few seconds)")
        Qt.QTimer.singleShot(0, self._expandTree)

    def _expandTree(self):
        tree = self.viewWidget()
        tree.expandAll()
        self.statusBar().showMessage("All items expanded!", 3000)

    def onExpanded(self):
        self.resizeColumns()

    def collapseAllTree(self):
        self.viewWidget().collapseAll()

    def expandSelectionTree(self):
        tree = self.viewWidget()
        index = tree.currentIndex()
        if index.isValid():
            tree.expand(index)

    def collapseSelectionTree(self):
        tree = self.viewWidget()
        index = tree.currentIndex()
        if index.isValid():
            tree.collapse(index)

    def resizeColumns(self):
        tree = self.viewWidget()
        model = tree.model()
        if model is None:
            return
        for c in range(model.columnCount()):
            tree.resizeColumnToContents(c)

    def goIntoTree(self):
        tree = self.viewWidget()
        index = tree.currentIndex()
        base_index = self._mapToSource(index)

        if not index.isValid():
            return

        # do not enter if the item doesn't have any children
        if base_index.internalPointer().childCount() == 0:
            return

        tree.setRootIndex(index)
        tree.setCurrentIndex(index.child(0, 0))
        self._updateToolBar()

    def goUpTree(self):
        tree = self.viewWidget()
        index = tree.rootIndex()
        if not index.isValid():
            return
        index_parent = index.parent()

        tree.setRootIndex(index_parent)
        tree.setCurrentIndex(index)
        self._updateToolBar()

    def goTopTree(self):
        tree = self.viewWidget()
        current_root = tree.rootIndex()
        p = current_root.parent()
        while p.isValid():
            p = p.parent()
        tree.setRootIndex(p)
        tree.setCurrentIndex(p)
        self._updateToolBar()

    def _updateToolBar(self, current=None, previous=None):
        if not self._with_navigation_bar:
            return
        tree = self.viewWidget()
        if current is None:
            current = tree.currentIndex()
        goInto = False

        base_current = self._mapToSource(current)
        if current.isValid():
            ip = base_current.internalPointer()
            if ip is not None:
                goInto = ip.childCount() > 0
        self._navigationToolBar._goIntoAction.setEnabled(goInto)
        self._expandBar._expandSelectionAction.setEnabled(goInto)
        self._expandBar._collapseSelectionAction.setEnabled(goInto)

        goUp = tree.rootIndex().isValid()
        self._navigationToolBar._goUpAction.setEnabled(goUp)
        self._navigationToolBar._goTopAction.setEnabled(goUp)

        index = tree.rootIndex()
        self._navigationToolBar._navigationWidget.updateSelection(index)
